# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, List, Optional, Tuple, Union

import cv2
import numpy as np
from mmcv import color_val

from mmpose.core import (apply_bugeye_effect, apply_sunglasses_effect,
                         imshow_bboxes, imshow_keypoints)
from mmpose.datasets import DatasetInfo
from ..utils import (FrameMessage, copy_and_paste, expand_and_clamp,
                     get_cached_file_path, get_eye_keypoint_ids,
                     get_face_keypoint_ids, get_wrist_keypoint_ids,
                     load_image_from_disk_or_url, screen_matting)
from .builder import NODES
from .frame_drawing_node import FrameDrawingNode

try:
    import psutil
    psutil_proc = psutil.Process()
except (ImportError, ModuleNotFoundError):
    psutil_proc = None


@NODES.register_module()
class PoseVisualizerNode(FrameDrawingNode):
    """Draw the bbox and keypoint detection results.

    Args:
        name (str, optional): The node name (also thread name).
        frame_buffer (str): The name of the input buffer.
        output_buffer (str|list): The name(s) of the output buffer(s).
        enable_key (str|int, optional): Set a hot-key to toggle enable/disable
            of the node. If an int value is given, it will be treated as an
            ascii code of a key. Please note:
                1. If enable_key is set, the bypass method need to be
                    overridden to define the node behavior when disabled
                2. Some hot-key has been use for particular use. For example:
                    'q', 'Q' and 27 are used for quit
            Default: None
        enable (bool): Default enable/disable status. Default: True.
        kpt_thr (float): The threshold of keypoint score. Default: 0.3.
        radius (int): The radius of keypoint. Default: 4.
        thickness (int): The thickness of skeleton. Default: 2.
        bbox_color (str|tuple|dict): If a single color (a str like 'green' or
            a tuple like (0, 255, 0)), it will used to draw the bbox.
            Optionally, a dict can be given as a map from class labels to
            colors.
    """

    default_bbox_color = {
        'person': (148, 139, 255),
        'cat': (255, 255, 0),
        'dog': (255, 255, 0),
    }

    def __init__(self,
                 name: str,
                 frame_buffer: str,
                 output_buffer: Union[str, List[str]],
                 enable_key: Optional[Union[str, int]] = None,
                 enable: bool = True,
                 kpt_thr: float = 0.3,
                 radius: int = 4,
                 thickness: int = 2,
                 bbox_color: Optional[Union[str, Tuple, Dict]] = None):

        super().__init__(name, frame_buffer, output_buffer, enable_key, enable)

        self.kpt_thr = kpt_thr
        self.radius = radius
        self.thickness = thickness
        if bbox_color is None:
            self.bbox_color = self.default_bbox_color
        elif isinstance(bbox_color, dict):
            self.bbox_color = {k: color_val(v) for k, v in bbox_color.items()}
        else:
            self.bbox_color = color_val(bbox_color)

    def draw(self, frame_msg):
        canvas = frame_msg.get_image()
        pose_results = frame_msg.get_pose_results()

        if not pose_results:
            return canvas

        for pose_result in frame_msg.get_pose_results():
            model_cfg = pose_result['model_cfg']
            dataset_info = DatasetInfo(model_cfg.dataset_info)

            # Extract bboxes and poses
            bbox_preds = []
            bbox_labels = []
            pose_preds = []
            for pred in pose_result['preds']:
                if 'bbox' in pred:
                    bbox_preds.append(pred['bbox'])
                    bbox_labels.append(pred.get('label', None))
                pose_preds.append(pred['keypoints'])

            # Get bbox colors
            if isinstance(self.bbox_color, dict):
                bbox_colors = [
                    self.bbox_color.get(label, (0, 255, 0))
                    for label in bbox_labels
                ]
            else:
                bbox_labels = self.bbox_color

            # Draw bboxes
            if bbox_preds:
                bboxes = np.vstack(bbox_preds)

                imshow_bboxes(
                    canvas,
                    bboxes,
                    labels=bbox_labels,
                    colors=bbox_colors,
                    text_color='white',
                    font_scale=0.5,
                    show=False)

            # Draw poses
            if pose_preds:
                imshow_keypoints(
                    canvas,
                    pose_preds,
                    skeleton=dataset_info.skeleton,
                    kpt_score_thr=0.3,
                    pose_kpt_color=dataset_info.pose_kpt_color,
                    pose_link_color=dataset_info.pose_link_color,
                    radius=self.radius,
                    thickness=self.thickness)

        return canvas


@NODES.register_module()
class SunglassesNode(FrameDrawingNode):

    def __init__(self,
                 name: str,
                 frame_buffer: str,
                 output_buffer: Union[str, List[str]],
                 enable_key: Optional[Union[str, int]] = None,
                 enable: bool = True,
                 src_img_path: Optional[str] = None):

        super().__init__(name, frame_buffer, output_buffer, enable_key, enable)

        if src_img_path is None:
            # The image attributes to:
            # https://www.vecteezy.com/free-vector/glass
            # Glass Vectors by Vecteezy
            src_img_path = 'demo/resources/sunglasses.jpg'
        self.src_img = load_image_from_disk_or_url(src_img_path)

    def draw(self, frame_msg):
        canvas = frame_msg.get_image()
        pose_results = frame_msg.get_pose_results()
        if not pose_results:
            return canvas
        for pose_result in pose_results:
            model_cfg = pose_result['model_cfg']
            preds = pose_result['preds']
            left_eye_idx, right_eye_idx = get_eye_keypoint_ids(model_cfg)

            canvas = apply_sunglasses_effect(canvas, preds, self.src_img,
                                             left_eye_idx, right_eye_idx)
        return canvas


@NODES.register_module()
class SpriteNode(FrameDrawingNode):

    def __init__(self,
                 name: str,
                 frame_buffer: str,
                 output_buffer: Union[str, List[str]],
                 enable_key: Optional[Union[str, int]] = None,
                 enable: bool = True,
                 src_img_path: Optional[str] = None):

        super().__init__(name, frame_buffer, output_buffer, enable_key, enable)

        if src_img_path is None:
            # Sprites of Touhou characters :)
            # Come from https://www.deviantart.com/shadowbendy/art/Touhou-rpg-maker-vx-Sprite-1-812746920  # noqa: E501
            src_img_path = (
                'https://user-images.githubusercontent.com/'
                '26739999/151532276-33f968d9-917f-45e3-8a99-ebde60be83bb.png')
        self.src_img = load_image_from_disk_or_url(
            src_img_path, cv2.IMREAD_UNCHANGED)[:144, :108]
        tmp = np.array(np.split(self.src_img, range(36, 144, 36), axis=0))
        tmp = np.array(np.split(tmp, range(36, 108, 36), axis=2))
        self.sprites = tmp
        self.pos = None
        self.anime_frame = 0

    def apply_sprite_effect(self,
                            img,
                            pose_results,
                            left_hand_index,
                            right_hand_index,
                            kpt_thr=0.5):
        """Apply sprite effect.

        Args:
            img (np.ndarray): Image data.
            pose_results (list[dict]): The pose estimation results containing:
                - "keypoints" ([K,3]): detection result in [x, y, score]
            left_hand_index (int): Keypoint index of left hand
            right_hand_index (int): Keypoint index of right hand
            kpt_thr (float): The score threshold of required keypoints.
        """

        hm, wm = self.sprites.shape[2:4]
        # anchor points in the sunglasses mask
        if self.pos is None:
            self.pos = [img.shape[0] // 2, img.shape[1] // 2]

        if len(pose_results) == 0:
            return img

        kpts = pose_results[0]['keypoints']

        if kpts[left_hand_index, 2] < kpt_thr and kpts[right_hand_index,
                                                       2] < kpt_thr:
            aim = self.pos
        else:
            kpt_lhand = kpts[left_hand_index, :2][::-1]
            kpt_rhand = kpts[right_hand_index, :2][::-1]

            def distance(a, b):
                return (a[0] - b[0])**2 + (a[1] - b[1])**2

            # Go to the nearest hand
            if distance(kpt_lhand, self.pos) < distance(kpt_rhand, self.pos):
                aim = kpt_lhand
            else:
                aim = kpt_rhand

        pos_thr = 15
        if aim[0] < self.pos[0] - pos_thr:
            # Go down
            sprite = self.sprites[self.anime_frame][3]
            self.pos[0] -= 1
        elif aim[0] > self.pos[0] + pos_thr:
            # Go up
            sprite = self.sprites[self.anime_frame][0]
            self.pos[0] += 1
        elif aim[1] < self.pos[1] - pos_thr:
            # Go right
            sprite = self.sprites[self.anime_frame][1]
            self.pos[1] -= 1
        elif aim[1] > self.pos[1] + pos_thr:
            # Go left
            sprite = self.sprites[self.anime_frame][2]
            self.pos[1] += 1
        else:
            # Stay
            self.anime_frame = 0
            sprite = self.sprites[self.anime_frame][0]

        if self.anime_frame < 2:
            self.anime_frame += 1
        else:
            self.anime_frame = 0

        x = self.pos[0] - hm // 2
        y = self.pos[1] - wm // 2
        x = max(0, min(x, img.shape[0] - hm))
        y = max(0, min(y, img.shape[0] - wm))

        # Overlay image with transparent
        img[x:x + hm, y:y +
            wm] = (img[x:x + hm, y:y + wm] * (1 - sprite[:, :, 3:] / 255) +
                   sprite[:, :, :3] * (sprite[:, :, 3:] / 255)).astype('uint8')

        return img

    def draw(self, frame_msg):
        canvas = frame_msg.get_image()
        pose_results = frame_msg.get_pose_results()
        if not pose_results:
            return canvas
        for pose_result in pose_results:
            model_cfg = pose_result['model_cfg']
            preds = pose_result['preds']
            # left_hand_idx, right_hand_idx = get_wrist_keypoint_ids(model_cfg)  # noqa: E501
            left_hand_idx, right_hand_idx = get_eye_keypoint_ids(model_cfg)

            canvas = self.apply_sprite_effect(canvas, preds, left_hand_idx,
                                              right_hand_idx)
        return canvas


@NODES.register_module()
class BackgroundNode(FrameDrawingNode):

    def __init__(self,
                 name: str,
                 frame_buffer: str,
                 output_buffer: Union[str, List[str]],
                 enable_key: Optional[Union[str, int]] = None,
                 enable: bool = True,
                 src_img_path: Optional[str] = None,
                 cls_ids: Optional[List] = None,
                 cls_names: Optional[List] = None):

        super().__init__(name, frame_buffer, output_buffer, enable_key, enable)

        self.cls_ids = cls_ids
        self.cls_names = cls_names

        if src_img_path is None:
            src_img_path = 'https://user-images.githubusercontent.com/'\
                           '11788150/149731957-abd5c908-9c7f-45b2-b7bf-'\
                           '821ab30c6a3e.jpg'
        self.src_img = load_image_from_disk_or_url(src_img_path)

    def apply_background_effect(self,
                                img,
                                det_results,
                                background_img,
                                effect_region=(0.2, 0.2, 0.8, 0.8)):
        """Change background.

        Args:
            img (np.ndarray): Image data.
            det_results (list[dict]): The detection results containing:

                - "cls_id" (int): Class index.
                - "label" (str): Class label (e.g. 'person').
                - "bbox" (ndarray:(5, )): bounding box result
                    [x, y, w, h, score].
                - "mask" (ndarray:(w, h)): instance segmentation result.
            background_img (np.ndarray): Background image.
            effect_region (tuple(4, )): The region to apply mask,
                the coordinates are normalized (x1, y1, x2, y2).
        """
        if len(det_results) > 0:
            # Choose the one with the highest score.
            det_result = det_results[0]
            bbox = det_result['bbox']
            mask = det_result['mask'].astype(np.uint8)
            img = copy_and_paste(img, background_img, mask, bbox,
                                 effect_region)
            return img
        else:
            return background_img

    def draw(self, frame_msg):
        canvas = frame_msg.get_image()
        if canvas.shape != self.src_img.shape:
            self.src_img = cv2.resize(self.src_img, canvas.shape[:2])
        det_results = frame_msg.get_detection_results()
        if not det_results:
            return canvas

        full_preds = []
        for det_result in det_results:
            preds = det_result['preds']
            if self.cls_ids:
                # Filter results by class ID
                filtered_preds = [
                    p for p in preds if p['cls_id'] in self.cls_ids
                ]
            elif self.cls_names:
                # Filter results by class name
                filtered_preds = [
                    p for p in preds if p['label'] in self.cls_names
                ]
            else:
                filtered_preds = preds
            full_preds.extend(filtered_preds)

        canvas = self.apply_background_effect(canvas, full_preds, self.src_img)

        return canvas


@NODES.register_module()
class SaiyanNode(FrameDrawingNode):

    def __init__(self,
                 name: str,
                 frame_buffer: str,
                 output_buffer: Union[str, List[str]],
                 enable_key: Optional[Union[str, int]] = None,
                 enable: bool = True,
                 hair_img_path: Optional[str] = None,
                 light_video_path: Optional[str] = None,
                 cls_ids: Optional[List] = None,
                 cls_names: Optional[List] = None):

        super().__init__(name, frame_buffer, output_buffer, enable_key, enable)

        self.cls_ids = cls_ids
        self.cls_names = cls_names

        if hair_img_path is None:
            hair_img_path = 'https://user-images.githubusercontent.com/'\
                            '11788150/149732117-fcd2d804-dc2c-426c-bee7-'\
                            '94be6146e05c.png'
        self.hair_img = load_image_from_disk_or_url(hair_img_path)

        if light_video_path is None:
            light_video_path = get_cached_file_path(
                'https://'
                'user-images.githubusercontent.com/11788150/149732080'
                '-ea6cfeda-0dc5-4bbb-892a-3831e5580520.mp4')
        self.light_video_path = light_video_path
        self.light_video = cv2.VideoCapture(self.light_video_path)

    def apply_saiyan_effect(self,
                            img,
                            pose_results,
                            saiyan_img,
                            light_frame,
                            face_indices,
                            bbox_thr=0.3,
                            kpt_thr=0.5):
        """Apply saiyan hair effect.

        Args:
            img (np.ndarray): Image data.
            pose_results (list[dict]): The pose estimation results containing:
                - "keypoints" ([K,3]): keypoint detection result
                    in [x, y, score]
            saiyan_img (np.ndarray): Saiyan image with transparent background.
            light_frame (np.ndarray): Light image with green screen.
            face_indices (int): Keypoint index of the face
            kpt_thr (float): The score threshold of required keypoints.
        """
        img = img.copy()
        im_shape = img.shape
        # Apply lightning effects.
        light_mask = screen_matting(light_frame, color='green')

        # anchor points in the mask
        pts_src = np.array(
            [
                [84, 398],  # face kpt 0
                [331, 393],  # face kpt 16
                [84, 145],
                [331, 140]
            ],
            dtype=np.float32)

        for pose in pose_results:
            bbox = pose['bbox']

            if bbox[-1] < bbox_thr:
                continue

            mask_inst = pose['mask']
            # cache
            fg = img[np.where(mask_inst)]

            bbox = expand_and_clamp(bbox[:4], im_shape, s=3.0)
            # Apply light effects between fg and bg
            img = copy_and_paste(
                light_frame,
                img,
                light_mask,
                effect_region=(bbox[0] / im_shape[1], bbox[1] / im_shape[0],
                               bbox[2] / im_shape[1], bbox[3] / im_shape[0]))
            # pop
            img[np.where(mask_inst)] = fg

            # Apply Saiyan hair effects
            kpts = pose['keypoints']
            if kpts[face_indices[0], 2] < kpt_thr or kpts[face_indices[16],
                                                          2] < kpt_thr:
                continue

            kpt_0 = kpts[face_indices[0], :2]
            kpt_16 = kpts[face_indices[16], :2]
            # orthogonal vector
            vo = (kpt_0 - kpt_16)[::-1] * [-1, 1]

            # anchor points in the image by eye positions
            pts_tar = np.vstack([kpt_0, kpt_16, kpt_0 + vo, kpt_16 + vo])

            h_mat, _ = cv2.findHomography(pts_src, pts_tar)
            patch = cv2.warpPerspective(
                saiyan_img,
                h_mat,
                dsize=(img.shape[1], img.shape[0]),
                borderValue=(0, 0, 0))
            mask_patch = cv2.cvtColor(patch, cv2.COLOR_BGR2GRAY)
            mask_patch = (mask_patch > 1).astype(np.uint8)
            img = cv2.copyTo(patch, mask_patch, img)

        return img

    def draw(self, frame_msg):
        canvas = frame_msg.get_image()

        det_results = frame_msg.get_detection_results()
        if not det_results:
            return canvas

        pose_results = frame_msg.get_pose_results()
        if not pose_results:
            return canvas

        for pose_result in pose_results:
            model_cfg = pose_result['model_cfg']
            preds = pose_result['preds']
            face_indices = get_face_keypoint_ids(model_cfg)

            ret, frame = self.light_video.read()
            if not ret:
                self.light_video = cv2.VideoCapture(self.light_video_path)
                ret, frame = self.light_video.read()

            canvas = self.apply_saiyan_effect(canvas, preds, self.hair_img,
                                              frame, face_indices)

        return canvas


@NODES.register_module()
class MoustacheNode(FrameDrawingNode):

    def __init__(self,
                 name: str,
                 frame_buffer: str,
                 output_buffer: Union[str, List[str]],
                 enable_key: Optional[Union[str, int]] = None,
                 enable: bool = True,
                 src_img_path: Optional[str] = None):

        super().__init__(name, frame_buffer, output_buffer, enable_key, enable)

        if src_img_path is None:
            src_img_path = 'https://user-images.githubusercontent.com/'\
                           '11788150/149732141-3afbab55-252a-428c-b6d8'\
                           '-0e352f432651.jpeg'
        self.src_img = load_image_from_disk_or_url(src_img_path)

    def apply_moustache_effect(self,
                               img,
                               pose_results,
                               moustache_img,
                               face_indices,
                               kpt_thr=0.5):
        """Apply moustache effect.

        Args:
            img (np.ndarray): Image data.
            pose_results (list[dict]): The pose estimation results containing:
                - "keypoints" ([K,3]): keypoint detection result
                    in [x, y, score]
            moustache_img (np.ndarray): Moustache image with white background.
            left_eye_index (int): Keypoint index of left eye
            right_eye_index (int): Keypoint index of right eye
            kpt_thr (float): The score threshold of required keypoints.
        """

        hm, wm = moustache_img.shape[:2]
        # anchor points in the moustache mask
        pts_src = np.array([[1164, 741], [1729, 741], [1164, 1244],
                            [1729, 1244]],
                           dtype=np.float32)

        for pose in pose_results:
            kpts = pose['keypoints']
            if kpts[face_indices[32], 2] < kpt_thr \
                    or kpts[face_indices[34], 2] < kpt_thr \
                    or kpts[face_indices[61], 2] < kpt_thr \
                    or kpts[face_indices[63], 2] < kpt_thr:
                continue

            kpt_32 = kpts[face_indices[32], :2]
            kpt_34 = kpts[face_indices[34], :2]
            kpt_61 = kpts[face_indices[61], :2]
            kpt_63 = kpts[face_indices[63], :2]
            # anchor points in the image by eye positions
            pts_tar = np.vstack([kpt_32, kpt_34, kpt_61, kpt_63])

            h_mat, _ = cv2.findHomography(pts_src, pts_tar)
            patch = cv2.warpPerspective(
                moustache_img,
                h_mat,
                dsize=(img.shape[1], img.shape[0]),
                borderValue=(255, 255, 255))
            #  mask the white background area in the patch with a threshold 200
            mask = cv2.cvtColor(patch, cv2.COLOR_BGR2GRAY)
            mask = (mask < 200).astype(np.uint8)
            img = cv2.copyTo(patch, mask, img)

        return img

    def draw(self, frame_msg):
        canvas = frame_msg.get_image()
        pose_results = frame_msg.get_pose_results()
        if not pose_results:
            return canvas
        for pose_result in pose_results:
            model_cfg = pose_result['model_cfg']
            preds = pose_result['preds']
            face_indices = get_face_keypoint_ids(model_cfg)
            canvas = self.apply_moustache_effect(canvas, preds, self.src_img,
                                                 face_indices)
        return canvas


@NODES.register_module()
class BugEyeNode(FrameDrawingNode):

    def draw(self, frame_msg):
        canvas = frame_msg.get_image()
        pose_results = frame_msg.get_pose_results()
        if not pose_results:
            return canvas
        for pose_result in pose_results:
            model_cfg = pose_result['model_cfg']
            preds = pose_result['preds']
            left_eye_idx, right_eye_idx = get_eye_keypoint_ids(model_cfg)

            canvas = apply_bugeye_effect(canvas, preds, left_eye_idx,
                                         right_eye_idx)
        return canvas


@NODES.register_module()
class NoticeBoardNode(FrameDrawingNode):

    default_content_lines = ['This is a notice board!']

    def __init__(
        self,
        name: str,
        frame_buffer: str,
        output_buffer: Union[str, List[str]],
        enable_key: Optional[Union[str, int]] = None,
        enable: bool = True,
        content_lines: Optional[List[str]] = None,
        x_offset: int = 20,
        y_offset: int = 20,
        y_delta: int = 15,
        text_color: Union[str, Tuple[int, int, int]] = 'black',
        background_color: Union[str, Tuple[int, int, int]] = (255, 183, 0),
        text_scale: float = 0.4,
    ):
        super().__init__(name, frame_buffer, output_buffer, enable_key, enable)

        self.x_offset = x_offset
        self.y_offset = y_offset
        self.y_delta = y_delta
        self.text_color = color_val(text_color)
        self.background_color = color_val(background_color)
        self.text_scale = text_scale

        if content_lines:
            self.content_lines = content_lines
        else:
            self.content_lines = self.default_content_lines

    def draw(self, frame_msg: FrameMessage) -> np.ndarray:
        img = frame_msg.get_image()
        canvas = np.full(img.shape, self.background_color, dtype=img.dtype)

        x = self.x_offset
        y = self.y_offset

        max_len = max([len(line) for line in self.content_lines])

        def _put_line(line=''):
            nonlocal y
            cv2.putText(canvas, line, (x, y), cv2.FONT_HERSHEY_DUPLEX,
                        self.text_scale, self.text_color, 1)
            y += self.y_delta

        for line in self.content_lines:
            _put_line(line)

        x1 = max(0, self.x_offset)
        x2 = min(img.shape[1], int(x + max_len * self.text_scale * 20))
        y1 = max(0, self.y_offset - self.y_delta)
        y2 = min(img.shape[0], y)

        src1 = canvas[y1:y2, x1:x2]
        src2 = img[y1:y2, x1:x2]
        img[y1:y2, x1:x2] = cv2.addWeighted(src1, 0.5, src2, 0.5, 0)

        return img


@NODES.register_module()
class HatNode(FrameDrawingNode):

    def __init__(self,
                 name: str,
                 frame_buffer: str,
                 output_buffer: Union[str, List[str]],
                 enable_key: Optional[Union[str, int]] = None,
                 src_img_path: Optional[str] = None):

        super().__init__(name, frame_buffer, output_buffer, enable_key)

        if src_img_path is None:
            # The image attributes to:
            # http://616pic.com/sucai/1m9i70p52.html
            src_img_path = 'https://user-images.githubusercontent.' \
                           'com/28900607/149766271-2f591c19-9b67-4' \
                           'd92-8f94-c272396ca141.png'
        self.src_img = load_image_from_disk_or_url(src_img_path,
                                                   cv2.IMREAD_UNCHANGED)

    @staticmethod
    def apply_hat_effect(img,
                         pose_results,
                         hat_img,
                         left_eye_index,
                         right_eye_index,
                         kpt_thr=0.5):
        """Apply hat effect.
        Args:
            img (np.ndarray): Image data.
            pose_results (list[dict]): The pose estimation results containing:
                - "keypoints" ([K,3]): keypoint detection result in
                    [x, y, score]
            hat_img (np.ndarray): Hat image with white alpha channel.
            left_eye_index (int): Keypoint index of left eye
            right_eye_index (int): Keypoint index of right eye
            kpt_thr (float): The score threshold of required keypoints.
        """
        img_orig = img.copy()

        img = img_orig.copy()
        hm, wm = hat_img.shape[:2]
        # anchor points in the sunglasses mask
        a = 0.3
        b = 0.7
        pts_src = np.array([[a * wm, a * hm], [a * wm, b * hm],
                            [b * wm, a * hm], [b * wm, b * hm]],
                           dtype=np.float32)

        for pose in pose_results:
            kpts = pose['keypoints']

            if kpts[left_eye_index, 2] < kpt_thr or \
                    kpts[right_eye_index, 2] < kpt_thr:
                continue

            kpt_leye = kpts[left_eye_index, :2]
            kpt_reye = kpts[right_eye_index, :2]
            # orthogonal vector to the left-to-right eyes
            vo = 0.5 * (kpt_reye - kpt_leye)[::-1] * [-1, 1]
            veye = 0.5 * (kpt_reye - kpt_leye)

            # anchor points in the image by eye positions
            pts_tar = np.vstack([
                kpt_reye + 1 * veye + 5 * vo, kpt_reye + 1 * veye + 1 * vo,
                kpt_leye - 1 * veye + 5 * vo, kpt_leye - 1 * veye + 1 * vo
            ])

            h_mat, _ = cv2.findHomography(pts_src, pts_tar)
            patch = cv2.warpPerspective(
                hat_img,
                h_mat,
                dsize=(img.shape[1], img.shape[0]),
                borderValue=(255, 255, 255))
            #  mask the white background area in the patch with a threshold 200
            mask = (patch[:, :, -1] > 128)
            patch = patch[:, :, :-1]
            mask = mask * (cv2.cvtColor(patch, cv2.COLOR_BGR2GRAY) > 30)
            mask = mask.astype(np.uint8)

            img = cv2.copyTo(patch, mask, img)
        return img

    def draw(self, frame_msg):
        canvas = frame_msg.get_image()
        pose_results = frame_msg.get_pose_results()
        if not pose_results:
            return canvas
        for pose_result in pose_results:
            model_cfg = pose_result['model_cfg']
            preds = pose_result['preds']
            left_eye_idx, right_eye_idx = get_eye_keypoint_ids(model_cfg)

            canvas = self.apply_hat_effect(canvas, preds, self.src_img,
                                           left_eye_idx, right_eye_idx)
        return canvas


@NODES.register_module()
class FirecrackerNode(FrameDrawingNode):

    def __init__(self,
                 name: str,
                 frame_buffer: str,
                 output_buffer: Union[str, List[str]],
                 enable_key: Optional[Union[str, int]] = None,
                 src_img_path: Optional[str] = None):

        super().__init__(name, frame_buffer, output_buffer, enable_key)

        if src_img_path is None:
            self.src_img_path = 'https://user-images.githubusercontent' \
                                '.com/28900607/149766281-6376055c-ed8b' \
                                '-472b-991f-60e6ae6ee1da.gif'
        src_img = cv2.VideoCapture(self.src_img_path)

        self.frame_list = []
        ret, frame = src_img.read()
        while frame is not None:
            self.frame_list.append(frame)
            ret, frame = src_img.read()
        self.num_frames = len(self.frame_list)
        self.frame_idx = 0
        self.frame_period = 4  # each frame in gif lasts for 4 frames in video

    @staticmethod
    def apply_firecracker_effect(img,
                                 pose_results,
                                 firecracker_img,
                                 left_wrist_idx,
                                 right_wrist_idx,
                                 kpt_thr=0.5):
        """Apply firecracker effect.
        Args:
            img (np.ndarray): Image data.
            pose_results (list[dict]): The pose estimation results containing:
                - "keypoints" ([K,3]): keypoint detection result in
                    [x, y, score]
            firecracker_img (np.ndarray): Firecracker image with white
                background.
            left_wrist_idx (int): Keypoint index of left wrist
            right_wrist_idx (int): Keypoint index of right wrist
            kpt_thr (float): The score threshold of required keypoints.
        """

        hm, wm = firecracker_img.shape[:2]
        # anchor points in the firecracker mask
        pts_src = np.array([[0. * wm, 0. * hm], [0. * wm, 1. * hm],
                            [1. * wm, 0. * hm], [1. * wm, 1. * hm]],
                           dtype=np.float32)

        h, w = img.shape[:2]
        h_tar = h / 3
        w_tar = h_tar / hm * wm

        for pose in pose_results:
            kpts = pose['keypoints']

            if kpts[left_wrist_idx, 2] > kpt_thr:
                kpt_lwrist = kpts[left_wrist_idx, :2]
                # anchor points in the image by eye positions
                pts_tar = np.vstack([
                    kpt_lwrist - [w_tar / 2, 0],
                    kpt_lwrist - [w_tar / 2, -h_tar],
                    kpt_lwrist + [w_tar / 2, 0],
                    kpt_lwrist + [w_tar / 2, h_tar]
                ])

                h_mat, _ = cv2.findHomography(pts_src, pts_tar)
                patch = cv2.warpPerspective(
                    firecracker_img,
                    h_mat,
                    dsize=(img.shape[1], img.shape[0]),
                    borderValue=(255, 255, 255))
                #  mask the white background area in the patch with
                # a threshold 200
                mask = cv2.cvtColor(patch, cv2.COLOR_BGR2GRAY)
                mask = (mask < 240).astype(np.uint8)
                img = cv2.copyTo(patch, mask, img)

            if kpts[right_wrist_idx, 2] > kpt_thr:
                kpt_rwrist = kpts[right_wrist_idx, :2]

                # anchor points in the image by eye positions
                pts_tar = np.vstack([
                    kpt_rwrist - [w_tar / 2, 0],
                    kpt_rwrist - [w_tar / 2, -h_tar],
                    kpt_rwrist + [w_tar / 2, 0],
                    kpt_rwrist + [w_tar / 2, h_tar]
                ])

                h_mat, _ = cv2.findHomography(pts_src, pts_tar)
                patch = cv2.warpPerspective(
                    firecracker_img,
                    h_mat,
                    dsize=(img.shape[1], img.shape[0]),
                    borderValue=(255, 255, 255))
                #  mask the white background area in the patch with
                # a threshold 200
                mask = cv2.cvtColor(patch, cv2.COLOR_BGR2GRAY)
                mask = (mask < 240).astype(np.uint8)
                img = cv2.copyTo(patch, mask, img)

        return img

    def draw(self, frame_msg):
        canvas = frame_msg.get_image()
        pose_results = frame_msg.get_pose_results()
        if not pose_results:
            return canvas

        frame = self.frame_list[self.frame_idx // self.frame_period]
        for pose_result in pose_results:
            model_cfg = pose_result['model_cfg']
            preds = pose_result['preds']
            left_wrist_idx, right_wrist_idx = get_wrist_keypoint_ids(model_cfg)

            canvas = self.apply_firecracker_effect(canvas, preds, frame,
                                                   left_wrist_idx,
                                                   right_wrist_idx)
        self.frame_idx = (self.frame_idx + 1) % (
            self.num_frames * self.frame_period)

        return canvas



# Copyright (c) OpenMMLab. All rights reserved.
from .buffer import BufferManager
from .event import EventManager
from .message import FrameMessage, Message, VideoEndingMessage
from .misc import (ImageCapture, copy_and_paste, expand_and_clamp,
                   get_cached_file_path, is_image_file, limit_max_fps,
                   load_image_from_disk_or_url, screen_matting)
from .pose import (get_eye_keypoint_ids, get_face_keypoint_ids,
                   get_hand_keypoint_ids, get_mouth_keypoint_ids,
                   get_wrist_keypoint_ids)

__all__ = [
    'BufferManager',
    'EventManager',
    'FrameMessage',
    'Message',
    'limit_max_fps',
    'VideoEndingMessage',
    'load_image_from_disk_or_url',
    'get_cached_file_path',
    'screen_matting',
    'expand_and_clamp',
    'copy_and_paste',
    'is_image_file',
    'ImageCapture',
    'get_eye_keypoint_ids',
    'get_face_keypoint_ids',
    'get_wrist_keypoint_ids',
    'get_mouth_keypoint_ids',
    'get_hand_keypoint_ids',
]



# Copyright (c) OpenMMLab. All rights reserved.

from argparse import ArgumentParser

from mmcv import Config, DictAction
from webcam_apis import WebcamRunner


def parse_args():
    parser = ArgumentParser('Lauch webcam runner')
    parser.add_argument(
        '--config',
        type=str,
        default='tools/webcam/configs/meow_dwen_dwen/meow_dwen_dwen.py')

    parser.add_argument(
        '--cfg-options',
        nargs='+',
        action=DictAction,
        default={},
        help='override some settings in the used config, the key-value pair '
        'in xxx=yyy format will be merged into config file. For example, '
        "'--cfg-options runner.camera_id=1 runner.synchronous=True'")

    return parser.parse_args()


def launch():
    args = parse_args()
    cfg = Config.fromfile(args.config)
    cfg.merge_from_dict(args.cfg_options)

    runner = WebcamRunner(**cfg.runner)
    runner.run()


if __name__ == '__main__':
    launch()



# Copyright (c) OpenMMLab. All rights reserved.
import logging
import time
from queue import Full, Queue
from threading import Thread
from typing import List, Optional, Union

import cv2
import numpy as np
from mmcv import color_val

from mmpose.utils.timer import RunningAverage
from .builder import NODES
from .node import Node

try:
    import psutil
    psutil_proc = psutil.Process()
except (ImportError, ModuleNotFoundError):
    psutil_proc = None


@NODES.register_module()
class ModelResultBindingNode(Node):

    def __init__(self, name: str, frame_buffer: str, result_buffer: str,
                 output_buffer: Union[str, List[str]]):
        super().__init__(name=name, enable=True)
        self.synchronous = None

        # Cache the latest model result
        self.last_result_msg = None
        self.last_output_msg = None

        # Inference speed analysis
        self.frame_fps = RunningAverage(window=10)
        self.frame_lag = RunningAverage(window=10)
        self.result_fps = RunningAverage(window=10)
        self.result_lag = RunningAverage(window=10)

        # Register buffers
        # Note that essential buffers will be set in set_runner() because
        # it depends on the runner.synchronous attribute.
        self.register_input_buffer(result_buffer, 'result', essential=False)
        self.register_input_buffer(frame_buffer, 'frame', essential=False)
        self.register_output_buffer(output_buffer)

    def set_runner(self, runner):
        super().set_runner(runner)

        # Set synchronous according to the runner
        if runner.synchronous:
            self.synchronous = True
            essential_input = 'result'
        else:
            self.synchronous = False
            essential_input = 'frame'

        # Set essential input buffer according to the synchronous setting
        for buffer_info in self._input_buffers:
            if buffer_info.input_name == essential_input:
                buffer_info.essential = True

    def process(self, input_msgs):
        result_msg = input_msgs['result']

        # Update last result
        if result_msg is not None:
            # Update result FPS
            if self.last_result_msg is not None:
                self.result_fps.update(
                    1.0 /
                    (result_msg.timestamp - self.last_result_msg.timestamp))
            # Update inference latency
            self.result_lag.update(time.time() - result_msg.timestamp)
            # Update last inference result
            self.last_result_msg = result_msg

        if not self.synchronous:
            # Asynchronous mode: Bind the latest result with the current frame.
            frame_msg = input_msgs['frame']

            self.frame_lag.update(time.time() - frame_msg.timestamp)

            # Bind result to frame
            if self.last_result_msg is not None:
                frame_msg.set_full_results(
                    self.last_result_msg.get_full_results())
                frame_msg.merge_route_info(
                    self.last_result_msg.get_route_info())

            output_msg = frame_msg

        else:
            # Synchronous mode: Directly output the frame that the model result
            # was obtained from.
            self.frame_lag.update(time.time() - result_msg.timestamp)
            output_msg = result_msg

        # Update frame fps and lag
        if self.last_output_msg is not None:
            self.frame_lag.update(time.time() - output_msg.timestamp)
            self.frame_fps.update(
                1.0 / (output_msg.timestamp - self.last_output_msg.timestamp))
        self.last_output_msg = output_msg

        return output_msg

    def _get_node_info(self):
        info = super()._get_node_info()
        info['result_fps'] = self.result_fps.average()
        info['result_lag (ms)'] = self.result_lag.average() * 1000
        info['frame_fps'] = self.frame_fps.average()
        info['frame_lag (ms)'] = self.frame_lag.average() * 1000
        return info


@NODES.register_module()
class MonitorNode(Node):

    _default_ignore_items = ['timestamp']

    def __init__(self,
                 name: str,
                 frame_buffer: str,
                 output_buffer: Union[str, List[str]],
                 enable_key: Optional[Union[str, int]] = None,
                 enable: bool = False,
                 x_offset=20,
                 y_offset=20,
                 y_delta=15,
                 text_color='black',
                 background_color=(255, 183, 0),
                 text_scale=0.4,
                 ignore_items: Optional[List[str]] = None):
        super().__init__(name=name, enable_key=enable_key, enable=enable)

        self.x_offset = x_offset
        self.y_offset = y_offset
        self.y_delta = y_delta
        self.text_color = color_val(text_color)
        self.background_color = color_val(background_color)
        self.text_scale = text_scale
        if ignore_items is None:
            self.ignore_items = self._default_ignore_items
        else:
            self.ignore_items = ignore_items

        self.register_input_buffer(frame_buffer, 'frame', essential=True)
        self.register_output_buffer(output_buffer)

    def process(self, input_msgs):
        frame_msg = input_msgs['frame']

        frame_msg.update_route_info(
            node_name='System Info',
            node_type='dummy',
            info=self._get_system_info())

        img = frame_msg.get_image()
        route_info = frame_msg.get_route_info()
        img = self._show_route_info(img, route_info)

        frame_msg.set_image(img)
        return frame_msg

    def _get_system_info(self):
        sys_info = {}
        if psutil_proc is not None:
            sys_info['CPU(%)'] = psutil_proc.cpu_percent()
            sys_info['Memory(%)'] = psutil_proc.memory_percent()
        return sys_info

    def _show_route_info(self, img, route_info):
        canvas = np.full(img.shape, self.background_color, dtype=img.dtype)

        x = self.x_offset
        y = self.y_offset

        max_len = 0

        def _put_line(line=''):
            nonlocal y, max_len
            cv2.putText(canvas, line, (x, y), cv2.FONT_HERSHEY_DUPLEX,
                        self.text_scale, self.text_color, 1)
            y += self.y_delta
            max_len = max(max_len, len(line))

        for node_info in route_info:
            title = f'{node_info["node"]}({node_info["node_type"]})'
            _put_line(title)
            for k, v in node_info['info'].items():
                if k in self.ignore_items:
                    continue
                if isinstance(v, float):
                    v = f'{v:.1f}'
                _put_line(f'    {k}: {v}')

        x1 = max(0, self.x_offset)
        x2 = min(img.shape[1], int(x + max_len * self.text_scale * 20))
        y1 = max(0, self.y_offset - self.y_delta)
        y2 = min(img.shape[0], y)

        src1 = canvas[y1:y2, x1:x2]
        src2 = img[y1:y2, x1:x2]
        img[y1:y2, x1:x2] = cv2.addWeighted(src1, 0.5, src2, 0.5, 0)

        return img

    def bypass(self, input_msgs):
        return input_msgs['frame']


@NODES.register_module()
class RecorderNode(Node):
    """Record the frames into a local file."""

    def __init__(
        self,
        name: str,
        frame_buffer: str,
        output_buffer: Union[str, List[str]],
        out_video_file: str,
        out_video_fps: int = 30,
        out_video_codec: str = 'mp4v',
        buffer_size: int = 30,
    ):
        super().__init__(name=name, enable_key=None, enable=True)

        self.queue = Queue(maxsize=buffer_size)
        self.out_video_file = out_video_file
        self.out_video_fps = out_video_fps
        self.out_video_codec = out_video_codec
        self.vwriter = None

        # Register buffers
        self.register_input_buffer(frame_buffer, 'frame', essential=True)
        self.register_output_buffer(output_buffer)

        # Start a new thread to write frame
        self.t_record = Thread(target=self._record, args=(), daemon=True)
        self.t_record.start()

    def process(self, input_msgs):

        frame_msg = input_msgs['frame']
        img = frame_msg.get_image() if frame_msg is not None else None
        img_queued = False

        while not img_queued:
            try:
                self.queue.put(img, timeout=1)
                img_queued = True
                logging.info(f'{self.name}: recorder received one frame!')
            except Full:
                logging.info(f'{self.name}: recorder jamed!')

        return frame_msg

    def _record(self):

        while True:

            img = self.queue.get()

            if img is None:
                break

            if self.vwriter is None:
                fourcc = cv2.VideoWriter_fourcc(*self.out_video_codec)
                fps = self.out_video_fps
                frame_size = (img.shape[1], img.shape[0])
                self.vwriter = cv2.VideoWriter(self.out_video_file, fourcc,
                                               fps, frame_size)
                assert self.vwriter.isOpened()

            self.vwriter.write(img)

        logging.info('Video recorder released!')
        if self.vwriter is not None:
            self.vwriter.release()

    def on_exit(self):
        try:
            # Try putting a None into the output queue so the self.vwriter will
            # be released after all queue frames have been written to file.
            self.queue.put(None, timeout=1)
            self.t_record.join(timeout=1)
        except Full:
            pass

        if self.t_record.is_alive():
            # Force to release self.vwriter
            logging.info('Video recorder forced release!')
            if self.vwriter is not None:
                self.vwriter.release()



# Copyright (c) OpenMMLab. All rights reserved.
from .dist_utils import allreduce_grads, sync_random_seed
from .regularizations import WeightNormClipHook

__all__ = ['allreduce_grads', 'WeightNormClipHook', 'sync_random_seed']



# Copyright (c) OpenMMLab. All rights reserved.
runner = dict(
    name='Debug CamRunner',
    camera_id=0,
    camera_fps=20,
    nodes=[
        dict(
            type='MonitorNode',
            name='Monitor',
            enable_key='m',
            frame_buffer='_frame_',
            output_buffer='display'),
        dict(
            type='RecorderNode',
            name='Recorder',
            out_video_file='webcam_output.mp4',
            frame_buffer='display',
            output_buffer='_display_')
    ])



# Copyright (c) OpenMMLab. All rights reserved.
import os.path as osp
import tempfile
import warnings
from collections import OrderedDict

import json_tricks as json
import numpy as np
from mmcv import Config, deprecated_api_warning

from ...builder import DATASETS
from ..base import Kpt2dSviewRgbImgTopDownDataset


@DATASETS.register_module()
class TopDownH36MDataset(Kpt2dSviewRgbImgTopDownDataset):
    """Human3.6M dataset for top-down 2D pose estimation.

    "Human3.6M: Large Scale Datasets and Predictive Methods for 3D Human
    Sensing in Natural Environments", TPAMI`2014.
    More details can be found in the `paper
    <http://vision.imar.ro/human3.6m/pami-h36m.pdf>`__.

    Human3.6M keypoint indexes::

        0: 'root (pelvis)',
        1: 'right_hip',
        2: 'right_knee',
        3: 'right_foot',
        4: 'left_hip',
        5: 'left_knee',
        6: 'left_foot',
        7: 'spine',
        8: 'thorax',
        9: 'neck_base',
        10: 'head',
        11: 'left_shoulder',
        12: 'left_elbow',
        13: 'left_wrist',
        14: 'right_shoulder',
        15: 'right_elbow',
        16: 'right_wrist'

    Args:
        ann_file (str): Path to the annotation file.
        img_prefix (str): Path to a directory where images are held.
            Default: None.
        data_cfg (dict): config
        pipeline (list[dict | callable]): A sequence of data transforms.
        dataset_info (DatasetInfo): A class containing all dataset info.
        test_mode (bool): Store True when building test or
            validation dataset. Default: False.
    """

    def __init__(self,
                 ann_file,
                 img_prefix,
                 data_cfg,
                 pipeline,
                 dataset_info=None,
                 test_mode=False):

        if dataset_info is None:
            warnings.warn(
                'dataset_info is missing. '
                'Check https://github.com/open-mmlab/mmpose/pull/663 '
                'for details.', DeprecationWarning)
            cfg = Config.fromfile('configs/_base_/datasets/h36m.py')
            dataset_info = cfg._cfg_dict['dataset_info']

        super().__init__(
            ann_file,
            img_prefix,
            data_cfg,
            pipeline,
            dataset_info=dataset_info,
            test_mode=test_mode)

        self.db = self._get_db()

        print(f'=> num_images: {self.num_images}')
        print(f'=> load {len(self.db)} samples')

    def _get_db(self):
        """Load dataset."""
        gt_db = []
        bbox_id = 0
        num_joints = self.ann_info['num_joints']
        for img_id in self.img_ids:

            ann_ids = self.coco.getAnnIds(imgIds=img_id, iscrowd=False)
            objs = self.coco.loadAnns(ann_ids)

            for obj in objs:
                if max(obj['keypoints']) == 0:
                    continue
                joints_3d = np.zeros((num_joints, 3), dtype=np.float32)
                joints_3d_visible = np.zeros((num_joints, 3), dtype=np.float32)

                keypoints = np.array(obj['keypoints']).reshape(-1, 3)
                joints_3d[:, :2] = keypoints[:, :2]
                joints_3d_visible[:, :2] = np.minimum(1, keypoints[:, 2:3])

                # use 1.25 padded bbox as input
                center, scale = self._xywh2cs(*obj['bbox'][:4])

                image_file = osp.join(self.img_prefix, self.id2name[img_id])

                gt_db.append({
                    'image_file': image_file,
                    'center': center,
                    'scale': scale,
                    'rotation': 0,
                    'joints_3d': joints_3d,
                    'joints_3d_visible': joints_3d_visible,
                    'dataset': self.dataset_name,
                    'bbox': obj['bbox'],
                    'bbox_score': 1,
                    'bbox_id': bbox_id
                })
                bbox_id = bbox_id + 1
        gt_db = sorted(gt_db, key=lambda x: x['bbox_id'])

        return gt_db

    @deprecated_api_warning(name_dict=dict(outputs='results'))
    def evaluate(self, results, res_folder=None, metric='PCK', **kwargs):
        """Evaluate human3.6m 2d keypoint results. The pose prediction results
        will be saved in `${res_folder}/result_keypoints.json`.

        Note:
            - batch_size: N
            - num_keypoints: K
            - heatmap height: H
            - heatmap width: W

        Args:
            results (list[dict]): Testing results containing the following
                items:

                - preds (np.ndarray[N,K,3]): The first two dimensions are
                    coordinates, score is the third dimension of the array.
                - boxes (np.ndarray[N,6]): [center[0], center[1], scale[0],
                    scale[1],area, score]
                - image_paths (list[str]): For example, ['data/coco/val2017
                    /000000393226.jpg']
                - heatmap (np.ndarray[N, K, H, W]): model output heatmap
                - bbox_id (list(int)).
            res_folder (str, optional): The folder to save the testing
                results. If not specified, a temp folder will be created.
                Default: None.
            metric (str | list[str]): Metric to be performed. Defaults: 'PCK'.

        Returns:
            dict: Evaluation results for evaluation metric.
        """
        metrics = metric if isinstance(metric, list) else [metric]
        allowed_metrics = ['PCK', 'EPE']
        for metric in metrics:
            if metric not in allowed_metrics:
                raise KeyError(f'metric {metric} is not supported')

        if res_folder is not None:
            tmp_folder = None
            res_file = osp.join(res_folder, 'result_keypoints.json')
        else:
            tmp_folder = tempfile.TemporaryDirectory()
            res_file = osp.join(tmp_folder.name, 'result_keypoints.json')

        kpts = []
        for result in results:
            preds = result['preds']
            boxes = result['boxes']
            image_paths = result['image_paths']
            bbox_ids = result['bbox_ids']

            batch_size = len(image_paths)
            for i in range(batch_size):
                image_id = self.name2id[image_paths[i][len(self.img_prefix):]]

                kpts.append({
                    'keypoints': preds[i].tolist(),
                    'center': boxes[i][0:2].tolist(),
                    'scale': boxes[i][2:4].tolist(),
                    'area': float(boxes[i][4]),
                    'score': float(boxes[i][5]),
                    'image_id': image_id,
                    'bbox_id': bbox_ids[i]
                })
        kpts = self._sort_and_unique_bboxes(kpts)

        self._write_keypoint_results(kpts, res_file)
        info_str = self._report_metric(res_file, metrics)
        name_value = OrderedDict(info_str)

        if tmp_folder is not None:
            tmp_folder.cleanup()

        return name_value

    @staticmethod
    def _write_keypoint_results(keypoints, res_file):
        """Write results into a json file."""

        with open(res_file, 'w') as f:
            json.dump(keypoints, f, sort_keys=True, indent=4)



# Copyright (c) OpenMMLab. All rights reserved.
runner = dict(
    # Basic configurations of the runner
    name='Super Saiyan Effects',
    camera_id=0,
    camera_fps=30,
    synchronous=False,
    # Define nodes.
    # The configuration of a node usually includes:
    #   1. 'type': Node class name
    #   2. 'name': Node name
    #   3. I/O buffers (e.g. 'input_buffer', 'output_buffer'): specify the
    #       input and output buffer names. This may depend on the node class.
    #   4. 'enable_key': assign a hot-key to toggle enable/disable this node.
    #       This may depend on the node class.
    #   5. Other class-specific arguments
    nodes=[
        # 'DetectorNode':
        # This node performs object detection from the frame image using an
        # MMDetection model.
        dict(
            type='DetectorNode',
            name='Detector',
            model_config='demo/mmdetection_cfg/mask_rcnn_r50_fpn_2x_coco.py',
            model_checkpoint='https://download.openmmlab.com/'
            'mmdetection/v2.0/mask_rcnn/mask_rcnn_r50_fpn_2x_coco/'
            'mask_rcnn_r50_fpn_2x_coco_bbox_mAP-0.392'
            '__segm_mAP-0.354_20200505_003907-3e542a40.pth',
            input_buffer='_input_',  # `_input_` is a runner-reserved buffer
            output_buffer='det_result'),
        # 'TopDownPoseEstimatorNode':
        # This node performs keypoint detection from the frame image using an
        # MMPose top-down model. Detection results is needed.
        dict(
            type='TopDownPoseEstimatorNode',
            name='Human Pose Estimator',
            model_config='configs/wholebody/2d_kpt_sview_rgb_img/'
            'topdown_heatmap/coco-wholebody/'
            'vipnas_mbv3_coco_wholebody_256x192_dark.py',
            model_checkpoint='https://openmmlab-share.oss-cn-hangz'
            'hou.aliyuncs.com/mmpose/top_down/vipnas/vipnas_mbv3_co'
            'co_wholebody_256x192_dark-e2158108_20211205.pth',
            cls_names=['person'],
            input_buffer='det_result',
            output_buffer='human_pose'),
        # 'ModelResultBindingNode':
        # This node binds the latest model inference result with the current
        # frame. (This means the frame image and inference result may be
        # asynchronous).
        dict(
            type='ModelResultBindingNode',
            name='ResultBinder',
            frame_buffer='_frame_',  # `_frame_` is a runner-reserved buffer
            result_buffer='human_pose',
            output_buffer='frame'),
        # 'SaiyanNode':
        # This node draw the Super Saiyan effect in the frame image.
        # Pose results is needed.
        dict(
            type='SaiyanNode',
            name='Visualizer',
            enable_key='s',
            cls_names=['person'],
            enable=True,
            frame_buffer='frame',
            output_buffer='vis_saiyan'),
        # 'NoticeBoardNode':
        # This node show a notice board with given content, e.g. help
        # information.
        dict(
            type='NoticeBoardNode',
            name='Helper',
            enable_key='h',
            frame_buffer='vis_saiyan',
            output_buffer='vis',
            content_lines=[
                'This is a demo for super saiyan effects. Have fun!', '',
                'Hot-keys:', '"s": Saiyan effect',
                '"h": Show help information',
                '"m": Show diagnostic information', '"q": Exit'
            ],
        ),
        # 'MonitorNode':
        # This node show diagnostic information in the frame image. It can
        # be used for debugging or monitoring system resource status.
        dict(
            type='MonitorNode',
            name='Monitor',
            enable_key='m',
            enable=False,
            frame_buffer='vis',
            output_buffer='_display_')  # `_frame_` is a runner-reserved buffer
    ])



# Copyright (c) OpenMMLab. All rights reserved.
from .builder import NODES
from .faceswap_node import FaceSwapNode
from .frame_effect_node import (BackgroundNode, BugEyeNode, MoustacheNode,
                                NoticeBoardNode, PoseVisualizerNode,
                                SaiyanNode, SunglassesNode)
from .helper_node import ModelResultBindingNode, MonitorNode, RecorderNode
from .mmdet_node import DetectorNode
from .mmpose_node import TopDownPoseEstimatorNode
from .valentinemagic_node import ValentineMagicNode
from .xdwendwen_node import XDwenDwenNode

__all__ = [
    'NODES', 'PoseVisualizerNode', 'DetectorNode', 'TopDownPoseEstimatorNode',
    'MonitorNode', 'BugEyeNode', 'SunglassesNode', 'ModelResultBindingNode',
    'NoticeBoardNode', 'RecorderNode', 'FaceSwapNode', 'MoustacheNode',
    'SaiyanNode', 'BackgroundNode', 'XDwenDwenNode', 'ValentineMagicNode'
]



_base_ = [
    '../../../../_base_/default_runtime.py',
    '../../../../_base_/datasets/mpii.py'
]
evaluation = dict(interval=10, metric='PCKh', save_best='PCKh')

optimizer = dict(
    type='Adam',
    lr=5e-4,
)
optimizer_config = dict(grad_clip=None)
# learning policy
lr_config = dict(
    policy='step',
    warmup='linear',
    warmup_iters=500,
    warmup_ratio=0.001,
    step=[170, 200])
total_epochs = 210
log_config = dict(
    interval=50, hooks=[
        dict(type='TextLoggerHook'),
    ])

target_type = 'GaussianHeatmap'
channel_cfg = dict(
    num_output_channels=16,
    dataset_joints=16,
    dataset_channel=list(range(16)),
    inference_channel=list(range(16)))

# model settings
model = dict(
    type='TopDown',
    pretrained='https://download.openmmlab.com/mmpose/'
    'pretrain_models/hrnet_w48-8ef0771d.pth',
    backbone=dict(
        type='HRNet',
        in_channels=3,
        extra=dict(
            stage1=dict(
                num_modules=1,
                num_branches=1,
                block='BOTTLENECK',
                num_blocks=(4, ),
                num_channels=(64, )),
            stage2=dict(
                num_modules=1,
                num_branches=2,
                block='BASIC',
                num_blocks=(4, 4),
                num_channels=(48, 96)),
            stage3=dict(
                num_modules=4,
                num_branches=3,
                block='BASIC',
                num_blocks=(4, 4, 4),
                num_channels=(48, 96, 192)),
            stage4=dict(
                num_modules=3,
                num_branches=4,
                block='BASIC',
                num_blocks=(4, 4, 4, 4),
                num_channels=(48, 96, 192, 384))),
    ),
    keypoint_head=dict(
        type='TopdownHeatmapSimpleHead',
        in_channels=48,
        out_channels=channel_cfg['num_output_channels'],
        num_deconv_layers=0,
        extra=dict(final_conv_kernel=1, ),
        loss_keypoint=dict(type='JointsMSELoss', use_target_weight=True)),
    train_cfg=dict(),
    test_cfg=dict(
        flip_test=True,
        post_process='default',
        shift_heatmap=False,
        target_type=target_type,
        modulate_kernel=11,
        use_udp=True))

data_cfg = dict(
    image_size=[256, 256],
    heatmap_size=[64, 64],
    num_output_channels=channel_cfg['num_output_channels'],
    num_joints=channel_cfg['dataset_joints'],
    dataset_channel=channel_cfg['dataset_channel'],
    inference_channel=channel_cfg['inference_channel'],
    use_gt_bbox=True,
    bbox_file=None,
)

train_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='TopDownRandomFlip', flip_prob=0.5),
    dict(
        type='TopDownGetRandomScaleRotation', rot_factor=40, scale_factor=0.5),
    dict(type='TopDownAffine', use_udp=True),
    dict(type='ToTensor'),
    dict(
        type='NormalizeTensor',
        mean=[0.485, 0.456, 0.406],
        std=[0.229, 0.224, 0.225]),
    dict(
        type='TopDownGenerateTarget',
        sigma=2,
        encoding='UDP',
        target_type=target_type),
    dict(
        type='Collect',
        keys=['img', 'target', 'target_weight'],
        meta_keys=[
            'image_file', 'joints_3d', 'joints_3d_visible', 'center', 'scale',
            'rotation', 'flip_pairs'
        ]),
]

val_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='TopDownAffine', use_udp=True),
    dict(type='ToTensor'),
    dict(
        type='NormalizeTensor',
        mean=[0.485, 0.456, 0.406],
        std=[0.229, 0.224, 0.225]),
    dict(
        type='Collect',
        keys=['img'],
        meta_keys=['image_file', 'center', 'scale', 'rotation', 'flip_pairs']),
]

test_pipeline = val_pipeline

data_root = 'data/mpii'
data = dict(
    samples_per_gpu=64,
    workers_per_gpu=2,
    val_dataloader=dict(samples_per_gpu=32),
    test_dataloader=dict(samples_per_gpu=32),
    train=dict(
        type='TopDownMpiiDataset',
        ann_file=f'{data_root}/annotations/mpii_train.json',
        img_prefix=f'{data_root}/images/',
        data_cfg=data_cfg,
        pipeline=train_pipeline,
        dataset_info={{_base_.dataset_info}}),
    val=dict(
        type='TopDownMpiiDataset',
        ann_file=f'{data_root}/annotations/mpii_val.json',
        img_prefix=f'{data_root}/images/',
        data_cfg=data_cfg,
        pipeline=val_pipeline,
        dataset_info={{_base_.dataset_info}}),
    test=dict(
        type='TopDownMpiiDataset',
        ann_file=f'{data_root}/annotations/mpii_val.json',
        img_prefix=f'{data_root}/images/',
        data_cfg=data_cfg,
        pipeline=test_pipeline,
        dataset_info={{_base_.dataset_info}}),
)



# Copyright (c) OpenMMLab. All rights reserved.
from dataclasses import dataclass
from typing import Dict, List, Optional, Union

from mmpose.apis import (get_track_id, inference_top_down_pose_model,
                         init_pose_model)
from mmpose.core import Smoother
from ..utils import Message
from .builder import NODES
from .node import Node


@dataclass
class TrackInfo:
    next_id: int = 0
    last_pose_preds: List = None


@NODES.register_module()
class TopDownPoseEstimatorNode(Node):

    def __init__(
            self,
            name: str,
            model_config: str,
            model_checkpoint: str,
            input_buffer: str,
            output_buffer: Union[str, List[str]],
            enable_key: Optional[Union[str, int]] = None,
            enable: bool = True,
            device: str = 'cuda:0',
            cls_ids: Optional[List] = None,
            cls_names: Optional[List] = None,
            bbox_thr: float = 0.5,
            smooth: bool = False,
            smooth_filter_cfg: str = 'configs/_base_/filters/one_euro.py'):
        super().__init__(name=name, enable_key=enable_key, enable=enable)

        # Init model
        self.model_config = model_config
        self.model_checkpoint = model_checkpoint
        self.device = device.lower()

        self.cls_ids = cls_ids
        self.cls_names = cls_names
        self.bbox_thr = bbox_thr

        if smooth:
            self.smoother = Smoother(smooth_filter_cfg, keypoint_dim=2)
        else:
            self.smoother = None
        # Init model
        self.model = init_pose_model(
            self.model_config,
            self.model_checkpoint,
            device=self.device.lower())

        # Store history for pose tracking
        self.track_info = TrackInfo()

        # Register buffers
        self.register_input_buffer(input_buffer, 'input', essential=True)
        self.register_output_buffer(output_buffer)

    def bypass(self, input_msgs):
        return input_msgs['input']

    def process(self, input_msgs: Dict[str, Message]) -> Message:

        input_msg = input_msgs['input']
        img = input_msg.get_image()
        det_results = input_msg.get_detection_results()

        if det_results is None:
            raise ValueError(
                'No detection results are found in the frame message.'
                f'{self.__class__.__name__} should be used after a '
                'detector node.')

        full_det_preds = []
        for det_result in det_results:
            det_preds = det_result['preds']
            if self.cls_ids:
                # Filter detection results by class ID
                det_preds = [
                    p for p in det_preds if p['cls_id'] in self.cls_ids
                ]
            elif self.cls_names:
                # Filter detection results by class name
                det_preds = [
                    p for p in det_preds if p['label'] in self.cls_names
                ]
            full_det_preds.extend(det_preds)

        # Inference pose
        pose_preds, _ = inference_top_down_pose_model(
            self.model,
            img,
            full_det_preds,
            bbox_thr=self.bbox_thr,
            format='xyxy')

        # Pose tracking
        pose_preds, next_id = get_track_id(
            pose_preds,
            self.track_info.last_pose_preds,
            self.track_info.next_id,
            use_oks=False,
            tracking_thr=0.3)

        self.track_info.next_id = next_id
        self.track_info.last_pose_preds = pose_preds.copy()

        # Pose smoothing
        if self.smoother:
            pose_preds = self.smoother.smooth(pose_preds)

        pose_result = {
            'preds': pose_preds,
            'model_cfg': self.model.cfg.copy(),
        }

        input_msg.add_pose_result(pose_result, tag=self.name)

        return input_msg



# Copyright (c) OpenMMLab. All rights reserved.
import time
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple, Union

import cv2
import numpy as np

from ..utils import (FrameMessage, get_eye_keypoint_ids, get_hand_keypoint_ids,
                     get_mouth_keypoint_ids, load_image_from_disk_or_url)
from .builder import NODES
from .frame_drawing_node import FrameDrawingNode


@dataclass
class HeartInfo():
    """Dataclass for heart information."""
    heart_type: int
    start_time: float
    start_pos: Tuple[int, int]
    end_pos: Tuple[int, int]


@NODES.register_module()
class ValentineMagicNode(FrameDrawingNode):

    def __init__(self,
                 name: str,
                 frame_buffer: str,
                 output_buffer: Union[str, List[str]],
                 enable_key: Optional[Union[str, int]] = None,
                 kpt_vis_thr: float = 0.3,
                 hand_heart_angle_thr: float = 90.0,
                 longest_duration: float = 2.0,
                 largest_ratio: float = 0.25,
                 hand_heart_img_path: Optional[str] = None,
                 flying_heart_img_path: Optional[str] = None,
                 hand_heart_dis_ratio_thr: float = 1.0,
                 flying_heart_dis_ratio_thr: float = 3.5,
                 num_persons: int = 2):

        super().__init__(
            name, frame_buffer, output_buffer, enable_key=enable_key)

        if hand_heart_img_path is None:
            hand_heart_img_path = 'https://user-images.githubusercontent.com/'\
                           '87690686/149731850-ea946766-a4e8-4efa-82f5'\
                           '-e2f0515db8ae.png'
        if flying_heart_img_path is None:
            flying_heart_img_path = 'https://user-images.githubusercontent.'\
                                    'com/87690686/153554948-937ce496-33dd-4'\
                                    '9ab-9829-0433fd7c13c4.png'

        self.hand_heart = load_image_from_disk_or_url(hand_heart_img_path)
        self.flying_heart = load_image_from_disk_or_url(flying_heart_img_path)

        self.kpt_vis_thr = kpt_vis_thr
        self.hand_heart_angle_thr = hand_heart_angle_thr
        self.hand_heart_dis_ratio_thr = hand_heart_dis_ratio_thr
        self.flying_heart_dis_ratio_thr = flying_heart_dis_ratio_thr
        self.longest_duration = longest_duration
        self.largest_ratio = largest_ratio
        self.num_persons = num_persons

        # record the heart infos for each person
        self.heart_infos = {}

    def _cal_distance(self, p1: np.ndarray, p2: np.ndarray) -> np.float64:
        """calculate the distance of points p1 and p2."""
        return np.sqrt((p1[0] - p2[0])**2 + (p1[1] - p2[1])**2)

    def _cal_angle(self, p1: np.ndarray, p2: np.ndarray, p3: np.ndarray,
                   p4: np.ndarray) -> np.float64:
        """calculate the angle of vectors v1(constructed by points p2 and p1)
        and v2(constructed by points p4 and p3)"""
        v1 = p2 - p1
        v2 = p4 - p3

        vector_prod = v1[0] * v2[0] + v1[1] * v2[1]
        length_prod = np.sqrt(pow(v1[0], 2) + pow(v1[1], 2)) * np.sqrt(
            pow(v2[0], 2) + pow(v2[1], 2))
        cos = vector_prod * 1.0 / (length_prod * 1.0 + 1e-6)

        return (np.arccos(cos) / np.pi) * 180

    def _check_heart(self, pred: Dict[str,
                                      np.ndarray], hand_indices: List[int],
                     mouth_index: int, eye_indices: List[int]) -> int:
        """Check the type of Valentine Magic based on the pose results and
        keypoint indices of hand, mouth. and eye.

        Args:
            pred(dict): The pose estimation results containing:
                - "keypoints" (np.ndarray[K,3]): keypoint detection result
                                                 in [x, y, score]
            hand_indices(list[int]): keypoint indices of hand
            mouth_index(int): keypoint index of mouth
            eye_indices(list[int]): keypoint indices of eyes

        Returns:
            int: a number representing the type of heart pose,
                 0: None, 1: hand heart, 2: left hand blow kiss,
                 3: right hand blow kiss
        """
        kpts = pred['keypoints']

        left_eye_idx, right_eye_idx = eye_indices
        left_eye_pos = kpts[left_eye_idx][:2]
        right_eye_pos = kpts[right_eye_idx][:2]
        eye_dis = self._cal_distance(left_eye_pos, right_eye_pos)

        # these indices are corresoponding to the following keypoints:
        # left_hand_root, left_pinky_finger1,
        # left_pinky_finger3, left_pinky_finger4,
        # right_hand_root, right_pinky_finger1
        # right_pinky_finger3, right_pinky_finger4

        both_hands_vis = True
        for i in [0, 17, 19, 20, 21, 38, 40, 41]:
            if kpts[hand_indices[i]][2] < self.kpt_vis_thr:
                both_hands_vis = False

        if both_hands_vis:
            p1 = kpts[hand_indices[20]][:2]
            p2 = kpts[hand_indices[19]][:2]
            p3 = kpts[hand_indices[17]][:2]
            p4 = kpts[hand_indices[0]][:2]
            left_angle = self._cal_angle(p1, p2, p3, p4)

            p1 = kpts[hand_indices[41]][:2]
            p2 = kpts[hand_indices[40]][:2]
            p3 = kpts[hand_indices[38]][:2]
            p4 = kpts[hand_indices[21]][:2]
            right_angle = self._cal_angle(p1, p2, p3, p4)

            hand_dis = self._cal_distance(kpts[hand_indices[20]][:2],
                                          kpts[hand_indices[41]][:2])

            if (left_angle < self.hand_heart_angle_thr
                    and right_angle < self.hand_heart_angle_thr
                    and hand_dis / eye_dis < self.hand_heart_dis_ratio_thr):
                return 1

        # these indices are corresoponding to the following keypoints:
        # left_middle_finger1, left_middle_finger4,
        left_hand_vis = True
        for i in [9, 12]:
            if kpts[hand_indices[i]][2] < self.kpt_vis_thr:
                left_hand_vis = False
                break
        # right_middle_finger1, right_middle_finger4

        right_hand_vis = True
        for i in [30, 33]:
            if kpts[hand_indices[i]][2] < self.kpt_vis_thr:
                right_hand_vis = False
                break

        mouth_vis = True
        if kpts[mouth_index][2] < self.kpt_vis_thr:
            mouth_vis = False

        if (not left_hand_vis and not right_hand_vis) or not mouth_vis:
            return 0

        mouth_pos = kpts[mouth_index]

        left_mid_hand_pos = (kpts[hand_indices[9]][:2] +
                             kpts[hand_indices[12]][:2]) / 2
        lefthand_mouth_dis = self._cal_distance(left_mid_hand_pos, mouth_pos)

        if lefthand_mouth_dis / eye_dis < self.flying_heart_dis_ratio_thr:
            return 2

        right_mid_hand_pos = (kpts[hand_indices[30]][:2] +
                              kpts[hand_indices[33]][:2]) / 2
        righthand_mouth_dis = self._cal_distance(right_mid_hand_pos, mouth_pos)

        if righthand_mouth_dis / eye_dis < self.flying_heart_dis_ratio_thr:
            return 3

        return 0

    def _get_heart_route(self, heart_type: int, cur_pred: Dict[str,
                                                               np.ndarray],
                         tar_pred: Dict[str,
                                        np.ndarray], hand_indices: List[int],
                         mouth_index: int) -> Tuple[int, int]:
        """get the start and end position of the heart, based on two keypoint
        results and keypoint indices of hand and mouth.

        Args:
            cur_pred(dict): The pose estimation results of current person,
                containing: the following keys:
                - "keypoints" (np.ndarray[K,3]): keypoint detection result
                                                 in [x, y, score]
            tar_pred(dict): The pose estimation results of target person,
                containing: the following keys:
                - "keypoints" (np.ndarray[K,3]): keypoint detection result
                                                 in [x, y, score]
            hand_indices(list[int]): keypoint indices of hand
            mouth_index(int): keypoint index of mouth

        Returns:
            tuple(int): the start position of heart
            tuple(int): the end position of heart
        """
        cur_kpts = cur_pred['keypoints']

        assert heart_type in [1, 2,
                              3], 'Can not determine the type of heart effect'

        if heart_type == 1:
            p1 = cur_kpts[hand_indices[20]][:2]
            p2 = cur_kpts[hand_indices[41]][:2]
        elif heart_type == 2:
            p1 = cur_kpts[hand_indices[9]][:2]
            p2 = cur_kpts[hand_indices[12]][:2]
        elif heart_type == 3:
            p1 = cur_kpts[hand_indices[30]][:2]
            p2 = cur_kpts[hand_indices[33]][:2]

        cur_x, cur_y = (p1 + p2) / 2
        # the mid point of two fingers
        start_pos = (int(cur_x), int(cur_y))

        tar_kpts = tar_pred['keypoints']
        end_pos = tar_kpts[mouth_index][:2]

        return start_pos, end_pos

    def _draw_heart(self, canvas: np.ndarray, heart_info: HeartInfo,
                    t_pass: float) -> np.ndarray:
        """draw the heart according to heart info and time."""
        start_x, start_y = heart_info.start_pos
        end_x, end_y = heart_info.end_pos

        scale = t_pass / self.longest_duration

        max_h, max_w = canvas.shape[:2]
        hm, wm = self.largest_ratio * max_h, self.largest_ratio * max_h
        new_h, new_w = int(hm * scale), int(wm * scale)

        x = int(start_x + scale * (end_x - start_x))
        y = int(start_y + scale * (end_y - start_y))

        y1 = max(0, y - int(new_h / 2))
        y2 = min(max_h - 1, y + int(new_h / 2))

        x1 = max(0, x - int(new_w / 2))
        x2 = min(max_w - 1, x + int(new_w / 2))

        target = canvas[y1:y2 + 1, x1:x2 + 1].copy()
        new_h, new_w = target.shape[:2]

        if new_h == 0 or new_w == 0:
            return canvas

        assert heart_info.heart_type in [
            1, 2, 3
        ], 'Can not determine the type of heart effect'
        if heart_info.heart_type == 1:  # hand heart
            patch = self.hand_heart.copy()
        elif heart_info.heart_type >= 2:  # hand blow kiss
            patch = self.flying_heart.copy()
            if heart_info.start_pos[0] > heart_info.end_pos[0]:
                patch = patch[:, ::-1]

        patch = cv2.resize(patch, (new_w, new_h))
        mask = cv2.cvtColor(patch, cv2.COLOR_BGR2GRAY)
        mask = (mask < 100)[..., None].astype(np.float32) * 0.8

        canvas[y1:y2 + 1, x1:x2 + 1] = patch * mask + target * (1 - mask)

        return canvas

    def draw(self, frame_msg: FrameMessage) -> np.ndarray:
        canvas = frame_msg.get_image()

        pose_results = frame_msg.get_pose_results()
        if not pose_results:
            return canvas

        for pose_result in pose_results:
            model_cfg = pose_result['model_cfg']

            preds = [pred.copy() for pred in pose_result['preds']]
            # if number of persons in the image is less than 2,
            # no heart effect will be triggered
            if len(preds) < self.num_persons:
                continue

            # if number of persons in the image is more than 2,
            # only use the first two pose results
            preds = preds[:self.num_persons]
            ids = [preds[i]['track_id'] for i in range(self.num_persons)]

            for id in self.heart_infos.copy():
                if id not in ids:
                    # if the id of a person not in previous heart_infos,
                    # delete the corresponding field
                    del self.heart_infos[id]

            for i in range(self.num_persons):
                id = preds[i]['track_id']

                # if the predicted person in previous heart_infos,
                # draw the heart
                if id in self.heart_infos.copy():
                    t_pass = time.time() - self.heart_infos[id].start_time

                    # the time passed since last heart pose less than
                    # longest_duration, continue to draw the heart
                    if t_pass < self.longest_duration:
                        canvas = self._draw_heart(canvas, self.heart_infos[id],
                                                  t_pass)
                    # reset corresponding heart info
                    else:
                        del self.heart_infos[id]
                else:
                    hand_indices = get_hand_keypoint_ids(model_cfg)
                    mouth_index = get_mouth_keypoint_ids(model_cfg)
                    eye_indices = get_eye_keypoint_ids(model_cfg)

                    # check the type of Valentine Magic based on pose results
                    # and keypoint indices of hand and mouth
                    heart_type = self._check_heart(preds[i], hand_indices,
                                                   mouth_index, eye_indices)
                    # trigger a Valentine Magic effect
                    if heart_type:
                        # get the route of heart
                        start_pos, end_pos = self._get_heart_route(
                            heart_type, preds[i],
                            preds[self.num_persons - 1 - i], hand_indices,
                            mouth_index)
                        start_time = time.time()
                        self.heart_infos[id] = HeartInfo(
                            heart_type, start_time, start_pos, end_pos)

        return canvas



# Copyright (c) OpenMMLab. All rights reserved.
runner = dict(
    # Basic configurations of the runner
    name='Pose Estimation',
    camera_id=0,
    camera_fps=20,
    synchronous=False,
    # Define nodes.
    # The configuration of a node usually includes:
    #   1. 'type': Node class name
    #   2. 'name': Node name
    #   3. I/O buffers (e.g. 'input_buffer', 'output_buffer'): specify the
    #       input and output buffer names. This may depend on the node class.
    #   4. 'enable_key': assign a hot-key to toggle enable/disable this node.
    #       This may depend on the node class.
    #   5. Other class-specific arguments
    nodes=[
        # 'DetectorNode':
        # This node performs object detection from the frame image using an
        # MMDetection model.
        dict(
            type='DetectorNode',
            name='Detector',
            model_config='demo/mmdetection_cfg/'
            'ssdlite_mobilenetv2_scratch_600e_coco.py',
            model_checkpoint='https://download.openmmlab.com'
            '/mmdetection/v2.0/ssd/'
            'ssdlite_mobilenetv2_scratch_600e_coco/ssdlite_mobilenetv2_'
            'scratch_600e_coco_20210629_110627-974d9307.pth',
            input_buffer='_input_',  # `_input_` is a runner-reserved buffer
            output_buffer='det_result'),
        # 'TopDownPoseEstimatorNode':
        # This node performs keypoint detection from the frame image using an
        # MMPose top-down model. Detection results is needed.
        dict(
            type='TopDownPoseEstimatorNode',
            name='Human Pose Estimator',
            model_config='configs/wholebody/2d_kpt_sview_rgb_img/'
            'topdown_heatmap/coco-wholebody/'
            'vipnas_mbv3_coco_wholebody_256x192_dark.py',
            model_checkpoint='https://openmmlab-share.oss-cn-hangz'
            'hou.aliyuncs.com/mmpose/top_down/vipnas/vipnas_mbv3_co'
            'co_wholebody_256x192_dark-e2158108_20211205.pth',
            cls_names=['person'],
            input_buffer='det_result',
            output_buffer='human_pose'),
        dict(
            type='TopDownPoseEstimatorNode',
            name='Animal Pose Estimator',
            model_config='configs/animal/2d_kpt_sview_rgb_img/topdown_heatmap'
            '/animalpose/hrnet_w32_animalpose_256x256.py',
            model_checkpoint='https://download.openmmlab.com/mmpose/animal/'
            'hrnet/hrnet_w32_animalpose_256x256-1aa7f075_20210426.pth',
            cls_names=['cat', 'dog', 'horse', 'sheep', 'cow'],
            input_buffer='human_pose',
            output_buffer='animal_pose'),
        # 'ModelResultBindingNode':
        # This node binds the latest model inference result with the current
        # frame. (This means the frame image and inference result may be
        # asynchronous).
        dict(
            type='ModelResultBindingNode',
            name='ResultBinder',
            frame_buffer='_frame_',  # `_frame_` is a runner-reserved buffer
            result_buffer='animal_pose',
            output_buffer='frame'),
        # 'HatNode':
        # This node draw the hat effect in the frame image.
        # Pose results is needed.
        dict(
            type='HatNode',
            name='Visualizer',
            enable_key='t',
            frame_buffer='frame',
            output_buffer='vis_hat'),
        # 'FirecrackerNode':
        # This node draw the firecracker effect in the frame image.
        # Pose results is needed.
        dict(
            type='FirecrackerNode',
            name='Visualizer',
            enable_key='f',
            frame_buffer='vis_hat',
            output_buffer='vis_firecracker'),
        # 'NoticeBoardNode':
        # This node show a notice board with given content, e.g. help
        # information.
        dict(
            type='NoticeBoardNode',
            name='Helper',
            enable_key='h',
            enable=True,
            frame_buffer='vis_firecracker',
            output_buffer='vis_notice',
            content_lines=[
                'This is a demo for pose visualization and simple image '
                'effects. Have fun!', '', 'Hot-keys:', '"t": Hat effect',
                '"f": Firecracker effect', '"h": Show help information',
                '"m": Show diagnostic information', '"q": Exit'
            ],
        ),
        # 'MonitorNode':
        # This node show diagnostic information in the frame image. It can
        # be used for debugging or monitoring system resource status.
        dict(
            type='MonitorNode',
            name='Monitor',
            enable_key='m',
            enable=False,
            frame_buffer='vis_notice',
            output_buffer='display'),
        # 'RecorderNode':
        # This node save the output video into a file.
        dict(
            type='RecorderNode',
            name='Recorder',
            out_video_file='record.mp4',
            frame_buffer='display',
            output_buffer='_display_'
            # `_display_` is a runner-reserved buffer
        )
    ])



# Copyright (c) OpenMMLab. All rights reserved.
import logging
import time
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from queue import Empty
from threading import Thread
from typing import Callable, Dict, List, Optional, Tuple, Union

from mmcv.utils.misc import is_method_overridden

from mmpose.utils import StopWatch
from ..utils import Message, VideoEndingMessage, limit_max_fps


@dataclass
class BufferInfo():
    """Dataclass for buffer information."""
    buffer_name: str
    input_name: Optional[str] = None
    essential: bool = False


@dataclass
class EventInfo():
    """Dataclass for event handler information."""
    event_name: str
    is_keyboard: bool = False
    handler_func: Optional[Callable] = None


class Node(Thread, metaclass=ABCMeta):
    """Base interface of functional module.

    Parameters:
        name (str, optional): The node name (also thread name).
        enable_key (str|int, optional): Set a hot-key to toggle enable/disable
            of the node. If an int value is given, it will be treated as an
            ascii code of a key. Please note:
                1. If enable_key is set, the bypass method need to be
                    overridden to define the node behavior when disabled
                2. Some hot-key has been use for particular use. For example:
                    'q', 'Q' and 27 are used for quit
            Default: None
        max_fps (int): Maximum FPS of the node. This is to avoid the node
            running unrestrictedly and causing large resource consuming.
            Default: 30
        input_check_interval (float): Minimum interval (in millisecond) between
            checking if input is ready. Default: 0.001
        enable (bool): Default enable/disable status. Default: True.
        daemon (bool): Whether node is a daemon. Default: True.
    """

    def __init__(self,
                 name: Optional[str] = None,
                 enable_key: Optional[Union[str, int]] = None,
                 max_fps: int = 30,
                 input_check_interval: float = 0.01,
                 enable: bool = True,
                 daemon=False):
        super().__init__(name=name, daemon=daemon)
        self._runner = None
        self._enabled = enable
        self.enable_key = enable_key
        self.max_fps = max_fps
        self.input_check_interval = input_check_interval

        # A partitioned buffer manager the runner's buffer manager that
        # only accesses the buffers related to the node
        self._buffer_manager = None

        # Input/output buffers are a list of registered buffers' information
        self._input_buffers = []
        self._output_buffers = []

        # Event manager is a copy of assigned runner's event manager
        self._event_manager = None

        # A list of registered event information
        # See register_event() for more information
        # Note that we recommend to handle events in nodes by registering
        # handlers, but one can still access the raw event by _event_manager
        self._registered_events = []

        # A list of (listener_threads, event_info)
        # See set_runner() for more information
        self._event_listener_threads = []

        # A timer to calculate node FPS
        self._timer = StopWatch(window=10)

        # Register enable toggle key
        if self.enable_key:
            # If the node allows toggling enable, it should override the
            # `bypass` method to define the node behavior when disabled.
            if not is_method_overridden('bypass', Node, self.__class__):
                raise NotImplementedError(
                    f'The node {self.__class__} does not support toggling'
                    'enable but got argument `enable_key`. To support toggling'
                    'enable, please override the `bypass` method of the node.')

            self.register_event(
                event_name=self.enable_key,
                is_keyboard=True,
                handler_func=self._toggle_enable,
            )

    @property
    def registered_buffers(self):
        return self._input_buffers + self._output_buffers

    @property
    def registered_events(self):
        return self._registered_events.copy()

    def _toggle_enable(self):
        self._enabled = not self._enabled

    def register_input_buffer(self,
                              buffer_name: str,
                              input_name: str,
                              essential: bool = False):
        """Register an input buffer, so that Node can automatically check if
        data is ready, fetch data from the buffers and format the inputs to
        feed into `process` method.

        This method can be invoked multiple times to register multiple input
        buffers.

        The subclass of Node should invoke `register_input_buffer` in its
        `__init__` method.

        Args:
            buffer_name (str): The name of the buffer
            input_name (str): The name of the fetched message from the
                corresponding buffer
            essential (bool): An essential input means the node will wait
                until the input is ready before processing. Otherwise, an
                inessential input will not block the processing, instead
                a None will be fetched if the buffer is not ready.
        """
        buffer_info = BufferInfo(buffer_name, input_name, essential)
        self._input_buffers.append(buffer_info)

    def register_output_buffer(self, buffer_name: Union[str, List[str]]):
        """Register one or multiple output buffers, so that the Node can
        automatically send the output of the `process` method to these buffers.

        The subclass of Node should invoke `register_output_buffer` in its
        `__init__` method.

        Args:
            buffer_name (str|list): The name(s) of the output buffer(s).
        """

        if not isinstance(buffer_name, list):
            buffer_name = [buffer_name]

        for name in buffer_name:
            buffer_info = BufferInfo(name)
            self._output_buffers.append(buffer_info)

    def register_event(self,
                       event_name: str,
                       is_keyboard: bool = False,
                       handler_func: Optional[Callable] = None):
        """Register an event. All events used in the node need to be registered
        in __init__(). If a callable handler is given, a thread will be create
        to listen and handle the event when the node starts.

        Args:
            Args:
            event_name (str|int): The event name. If is_keyboard==True,
                event_name should be a str (as char) or an int (as ascii)
            is_keyboard (bool): Indicate whether it is an keyboard
                event. If True, the argument event_name will be regarded as a
                key indicator.
            handler_func (callable, optional): The event handler function,
                which should be a collable object with no arguments or
                return values. Default: None.
        """
        event_info = EventInfo(event_name, is_keyboard, handler_func)
        self._registered_events.append(event_info)

    def set_runner(self, runner):
        # Get partitioned buffer manager
        buffer_names = [
            buffer.buffer_name
            for buffer in self._input_buffers + self._output_buffers
        ]
        self._buffer_manager = runner.buffer_manager.get_sub_manager(
            buffer_names)

        # Get event manager
        self._event_manager = runner.event_manager

    def _get_input_from_buffer(self) -> Tuple[bool, Optional[Dict]]:
        """Get and pack input data if it's ready. The function returns a tuple
        of a status flag and a packed data dictionary. If input_buffer is
        ready, the status flag will be True, and the packed data is a dict
        whose items are buffer names and corresponding messages (unready
        additional buffers will give a `None`). Otherwise, the status flag is
        False and the packed data is None.

        Returns:
            bool: status flag
            dict[str, Message]: the packed inputs where the key is the buffer
                name and the value is the Message got from the corresponding
                buffer.
        """
        buffer_manager = self._buffer_manager

        if buffer_manager is None:
            raise ValueError(f'{self.name}: Runner not set!')

        # Check that essential buffers are ready
        for buffer_info in self._input_buffers:
            if buffer_info.essential and buffer_manager.is_empty(
                    buffer_info.buffer_name):
                return False, None

        # Default input
        result = {
            buffer_info.input_name: None
            for buffer_info in self._input_buffers
        }

        for buffer_info in self._input_buffers:
            try:
                result[buffer_info.input_name] = buffer_manager.get(
                    buffer_info.buffer_name, block=False)
            except Empty:
                if buffer_info.essential:
                    # Return unsuccessful flag if any
                    # essential input is unready
                    return False, None

        return True, result

    def _send_output_to_buffers(self, output_msg):
        """Send output of the process method to registered output buffers.

        Args:
            output_msg (Message): output message
            force (bool, optional): If True, block until the output message
                has been put into all output buffers. Default: False
        """
        for buffer_info in self._output_buffers:
            buffer_name = buffer_info.buffer_name
            self._buffer_manager.put_force(buffer_name, output_msg)

    @abstractmethod
    def process(self, input_msgs: Dict[str, Message]) -> Union[Message, None]:
        """The core method that implement the function of the node. This method
        will be invoked when the node is enabled and the input data is ready.

        All subclasses of Node should override this method.

        Args:
            input_msgs (dict): The input data collected from the buffers. For
                each item, the key is the `input_name` of the registered input
                buffer, while the value is a Message instance fetched from the
                buffer (or None if the buffer is unessential and not ready).

        Returns:
            Message: The output message of the node. It will be send to all
                registered output buffers.
        """

    def bypass(self, input_msgs: Dict[str, Message]) -> Union[Message, None]:
        """The method that defines the node behavior when disabled. Note that
        if the node has an `enable_key`, this method should be override.

        The method input/output is same as it of `process` method.

        Args:
            input_msgs (dict): The input data collected from the buffers. For
                each item, the key is the `input_name` of the registered input
                buffer, while the value is a Message instance fetched from the
                buffer (or None if the buffer is unessential and not ready).

        Returns:
            Message: The output message of the node. It will be send to all
                registered output buffers.
        """
        raise NotImplementedError

    def _get_node_info(self):
        """Get route information of the node."""
        info = {'fps': self._timer.report('_FPS_'), 'timestamp': time.time()}
        return info

    def on_exit(self):
        """This method will be invoked on event `_exit_`.

        Subclasses should override this method to specifying the exiting
        behavior.
        """

    def run(self):
        """Method representing the Node's activity.

        This method override the standard run() method of Thread. Users should
        not override this method in subclasses.
        """

        logging.info(f'Node {self.name} starts')

        # Create event listener threads
        for event_info in self._registered_events:

            if event_info.handler_func is None:
                continue

            def event_listener():
                while True:
                    with self._event_manager.wait_and_handle(
                            event_info.event_name, event_info.is_keyboard):
                        event_info.handler_func()

            t_listener = Thread(target=event_listener, args=(), daemon=True)
            t_listener.start()
            self._event_listener_threads.append(t_listener)

        # Loop
        while True:
            # Exit
            if self._event_manager.is_set('_exit_'):
                self.on_exit()
                break

            # Check if input is ready
            input_status, input_msgs = self._get_input_from_buffer()

            # Input is not ready
            if not input_status:
                time.sleep(self.input_check_interval)
                continue

            # If a VideoEndingMessage is received, broadcast the signal
            # without invoking process() or bypass()
            video_ending = False
            for _, msg in input_msgs.items():
                if isinstance(msg, VideoEndingMessage):
                    self._send_output_to_buffers(msg)
                    video_ending = True
                    break

            if video_ending:
                self.on_exit()
                break

            # Check if enabled
            if not self._enabled:
                # Override bypass method to define node behavior when disabled
                output_msg = self.bypass(input_msgs)
            else:
                with self._timer.timeit():
                    with limit_max_fps(self.max_fps):
                        # Process
                        output_msg = self.process(input_msgs)

                if output_msg:
                    # Update route information
                    node_info = self._get_node_info()
                    output_msg.update_route_info(node=self, info=node_info)

            # Send output message
            if output_msg is not None:
                self._send_output_to_buffers(output_msg)

        logging.info(f'{self.name}: process ending.')



# Copyright (c) OpenMMLab. All rights reserved.
import time
import uuid
import warnings
from typing import Dict, List, Optional

import numpy as np


class Message():
    """Message base class.

    All message class should inherit this class. The basic use of a Message
    instance is to carray a piece of text message (self.msg) and a dict that
    stores structured data (self.data), e.g. frame image, model prediction,
    et al.

    A message may also hold route information, which is composed of
    information of all nodes the message has passed through.

    Parameters:
        msg (str): The text message.
        data (dict, optional): The structured data.
    """

    def __init__(self, msg: str = '', data: Optional[Dict] = None):
        self.msg = msg
        self.data = data if data else {}
        self.route_info = []
        self.timestamp = time.time()
        self.id = uuid.uuid4()

    def update_route_info(self,
                          node=None,
                          node_name: Optional[str] = None,
                          node_type: Optional[str] = None,
                          info: Optional[Dict] = None):
        """Append new node information to the route information.

        Args:
            node (Node, optional): An instance of Node that provides basic
                information like the node name and type. Default: None.
            node_name (str, optional): The node name. If node is given,
                node_name will be ignored. Default: None.
            node_type (str, optional): The class name of the node. If node
                is given, node_type will be ignored. Default: None.
            info (dict, optional): The node information, which is usually
                given by node.get_node_info(). Default: None.
        """
        if node is not None:
            if node_name is not None or node_type is not None:
                warnings.warn(
                    '`node_name` and `node_type` will be overridden if node'
                    'is provided.')
            node_name = node.name
            node_type = node.__class__.__name__

        node_info = {'node': node_name, 'node_type': node_type, 'info': info}
        self.route_info.append(node_info)

    def set_route_info(self, route_info: List):
        """Directly set the entire route information.

        Args:
            route_info (list): route information to set to the message.
        """
        self.route_info = route_info

    def merge_route_info(self, route_info: List):
        """Merge the given route information into the original one of the
        message. This is used for combining route information from multiple
        messages. The node information in the route will be reordered according
        to their timestamps.

        Args:
            route_info (list): route information to merge.
        """
        self.route_info += route_info
        self.route_info.sort(key=lambda x: x.get('timestamp', np.inf))

    def get_route_info(self) -> List:
        return self.route_info.copy()


class VideoEndingMessage(Message):
    """A special message to indicate the input video is ending."""


class FrameMessage(Message):
    """The message to store information of a video frame.

    A FrameMessage instance usually holds following data in self.data:
        - image (array): The frame image
        - detection_results (list): A list to hold detection results of
            multiple detectors. Each element is a tuple (tag, result)
        - pose_results (list): A list to hold pose estimation results of
            multiple pose estimator. Each element is a tuple (tag, result)
    """

    def __init__(self, img):
        super().__init__(data=dict(image=img))

    def get_image(self):
        """Get the frame image.

        Returns:
            array: The frame image.
        """
        return self.data.get('image', None)

    def set_image(self, img):
        """Set the frame image to the message."""
        self.data['image'] = img

    def add_detection_result(self, result, tag: str = None):
        """Add the detection result from one model into the message's
        detection_results.

        Args:
            tag (str, optional): Give a tag to the result, which can be used
                to retrieve specific results.
        """
        if 'detection_results' not in self.data:
            self.data['detection_results'] = []
        self.data['detection_results'].append((tag, result))

    def get_detection_results(self, tag: str = None):
        """Get detection results of the message.

        Args:
            tag (str, optional): If given, only the results with the tag
                will be retrieved. Otherwise all results will be retrieved.
                Default: None.

        Returns:
            list[dict]: The retrieved detection results
        """
        if 'detection_results' not in self.data:
            return None
        if tag is None:
            results = [res for _, res in self.data['detection_results']]
        else:
            results = [
                res for _tag, res in self.data['detection_results']
                if _tag == tag
            ]
        return results

    def add_pose_result(self, result, tag=None):
        """Add the pose estimation result from one model into the message's
        pose_results.

        Args:
            tag (str, optional): Give a tag to the result, which can be used
                to retrieve specific results.
        """
        if 'pose_results' not in self.data:
            self.data['pose_results'] = []
        self.data['pose_results'].append((tag, result))

    def get_pose_results(self, tag=None):
        """Get pose estimation results of the message.

        Args:
            tag (str, optional): If given, only the results with the tag
                will be retrieved. Otherwise all results will be retrieved.
                Default: None.

        Returns:
            list[dict]: The retrieved pose results
        """
        if 'pose_results' not in self.data:
            return None
        if tag is None:
            results = [res for _, res in self.data['pose_results']]
        else:
            results = [
                res for _tag, res in self.data['pose_results'] if _tag == tag
            ]
        return results

    def get_full_results(self):
        """Get all model predictions of the message.

        See set_full_results() for inference.

        Returns:
            dict: All model predictions, including:
                - detection_results
                - pose_results
        """
        result_keys = ['detection_results', 'pose_results']
        results = {k: self.data[k] for k in result_keys}
        return results

    def set_full_results(self, results):
        """Set full model results directly.

        Args:
            results (dict): All model predictions including:
                - detection_results (list): see also add_detection_results()
                - pose_results (list): see also add_pose_results()
        """
        self.data.update(results)



# Copyright (c) OpenMMLab. All rights reserved.
import argparse
import time
from collections import deque
from queue import Queue
from threading import Event, Lock, Thread

import cv2
import numpy as np

from mmpose.apis import (get_track_id, inference_top_down_pose_model,
                         init_pose_model, vis_pose_result)
from mmpose.core import Smoother, apply_bugeye_effect, apply_sunglasses_effect
from mmpose.utils import StopWatch

try:
    from mmdet.apis import inference_detector, init_detector
    has_mmdet = True
except (ImportError, ModuleNotFoundError):
    has_mmdet = False

try:
    import psutil
    psutil_proc = psutil.Process()
except (ImportError, ModuleNotFoundError):
    psutil_proc = None


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--cam-id', type=str, default='0')
    parser.add_argument(
        '--det-config',
        type=str,
        default='demo/mmdetection_cfg/'
        'ssdlite_mobilenetv2_scratch_600e_coco.py',
        help='Config file for detection')
    parser.add_argument(
        '--det-checkpoint',
        type=str,
        default='https://download.openmmlab.com/mmdetection/v2.0/ssd/'
        'ssdlite_mobilenetv2_scratch_600e_coco/ssdlite_mobilenetv2_'
        'scratch_600e_coco_20210629_110627-974d9307.pth',
        help='Checkpoint file for detection')
    parser.add_argument(
        '--enable-human-pose',
        type=int,
        default=1,
        help='Enable human pose estimation')
    parser.add_argument(
        '--enable-animal-pose',
        type=int,
        default=0,
        help='Enable animal pose estimation')
    parser.add_argument(
        '--human-pose-config',
        type=str,
        default='configs/wholebody/2d_kpt_sview_rgb_img/topdown_heatmap/'
        'coco-wholebody/vipnas_res50_coco_wholebody_256x192_dark.py',
        help='Config file for human pose')
    parser.add_argument(
        '--human-pose-checkpoint',
        type=str,
        default='https://download.openmmlab.com/'
        'mmpose/top_down/vipnas/'
        'vipnas_res50_wholebody_256x192_dark-67c0ce35_20211112.pth',
        help='Checkpoint file for human pose')
    parser.add_argument(
        '--human-det-ids',
        type=int,
        default=[1],
        nargs='+',
        help='Object category label of human in detection results.'
        'Default is [1(person)], following COCO definition.')
    parser.add_argument(
        '--animal-pose-config',
        type=str,
        default='configs/animal/2d_kpt_sview_rgb_img/topdown_heatmap/'
        'animalpose/hrnet_w32_animalpose_256x256.py',
        help='Config file for animal pose')
    parser.add_argument(
        '--animal-pose-checkpoint',
        type=str,
        default='https://download.openmmlab.com/mmpose/animal/hrnet/'
        'hrnet_w32_animalpose_256x256-1aa7f075_20210426.pth',
        help='Checkpoint file for animal pose')
    parser.add_argument(
        '--animal-det-ids',
        type=int,
        default=[16, 17, 18, 19, 20],
        nargs='+',
        help='Object category label of animals in detection results'
        'Default is [16(cat), 17(dog), 18(horse), 19(sheep), 20(cow)], '
        'following COCO definition.')
    parser.add_argument(
        '--device', default='cuda:0', help='Device used for inference')
    parser.add_argument(
        '--det-score-thr',
        type=float,
        default=0.5,
        help='bbox score threshold')
    parser.add_argument(
        '--kpt-thr', type=float, default=0.3, help='bbox score threshold')
    parser.add_argument(
        '--vis-mode',
        type=int,
        default=2,
        help='0-none. 1-detection only. 2-detection and pose.')
    parser.add_argument(
        '--sunglasses', action='store_true', help='Apply `sunglasses` effect.')
    parser.add_argument(
        '--bugeye', action='store_true', help='Apply `bug-eye` effect.')

    parser.add_argument(
        '--out-video-file',
        type=str,
        default=None,
        help='Record the video into a file. This may reduce the frame rate')

    parser.add_argument(
        '--out-video-fps',
        type=int,
        default=20,
        help='Set the FPS of the output video file.')

    parser.add_argument(
        '--buffer-size',
        type=int,
        default=-1,
        help='Frame buffer size. If set -1, the buffer size will be '
        'automatically inferred from the display delay time. Default: -1')

    parser.add_argument(
        '--inference-fps',
        type=int,
        default=10,
        help='Maximum inference FPS. This is to limit the resource consuming '
        'especially when the detection and pose model are lightweight and '
        'very fast. Default: 10.')

    parser.add_argument(
        '--display-delay',
        type=int,
        default=0,
        help='Delay the output video in milliseconds. This can be used to '
        'align the output video and inference results. The delay can be '
        'disabled by setting a non-positive delay time. Default: 0')

    parser.add_argument(
        '--synchronous-mode',
        action='store_true',
        help='Enable synchronous mode that video I/O and inference will be '
        'temporally aligned. Note that this will reduce the display FPS.')

    parser.add_argument(
        '--smooth',
        action='store_true',
        help='Apply a temporal filter to smooth the pose estimation results. '
        'See also --smooth-filter-cfg.')
    parser.add_argument(
        '--smooth-filter-cfg',
        type=str,
        default='configs/_base_/filters/one_euro.py',
        help='Config file of the filter to smooth the pose estimation '
        'results. See also --smooth.')

    return parser.parse_args()


def process_mmdet_results(mmdet_results, class_names=None, cat_ids=1):
    """Process mmdet results to mmpose input format.

    Args:
        mmdet_results: raw output of mmdet model
        class_names: class names of mmdet model
        cat_ids (int or List[int]): category id list that will be preserved
    Returns:
        List[Dict]: detection results for mmpose input
    """
    if isinstance(mmdet_results, tuple):
        mmdet_results = mmdet_results[0]

    if isinstance(class_names, str):
        class_names = (class_names, )

    if not isinstance(cat_ids, (list, tuple)):
        cat_ids = [cat_ids]

    # only keep bboxes of interested classes
    bbox_results = [mmdet_results[i - 1] for i in cat_ids]
    bboxes = np.vstack(bbox_results)

    # get textual labels of classes
    labels = np.concatenate([
        np.full(bbox.shape[0], i - 1, dtype=np.int32)
        for i, bbox in zip(cat_ids, bbox_results)
    ])
    if class_names is None:
        labels = [f'class: {i}' for i in labels]
    else:
        labels = [class_names[i] for i in labels]

    det_results = []
    for bbox, label in zip(bboxes, labels):
        det_result = dict(bbox=bbox, label=label)
        det_results.append(det_result)
    return det_results


def read_camera():
    # init video reader
    print('Thread "input" started')
    cam_id = args.cam_id
    if cam_id.isdigit():
        cam_id = int(cam_id)
    vid_cap = cv2.VideoCapture(cam_id)
    if not vid_cap.isOpened():
        print(f'Cannot open camera (ID={cam_id})')
        exit()

    while not event_exit.is_set():
        # capture a camera frame
        ret_val, frame = vid_cap.read()
        if ret_val:
            ts_input = time.time()

            event_inference_done.clear()
            with input_queue_mutex:
                input_queue.append((ts_input, frame))

            if args.synchronous_mode:
                event_inference_done.wait()

            frame_buffer.put((ts_input, frame))
        else:
            # input ending signal
            frame_buffer.put((None, None))
            break

    vid_cap.release()


def inference_detection():
    print('Thread "det" started')
    stop_watch = StopWatch(window=10)
    min_interval = 1.0 / args.inference_fps
    _ts_last = None  # timestamp when last inference was done

    while True:
        while len(input_queue) < 1:
            time.sleep(0.001)
        with input_queue_mutex:
            ts_input, frame = input_queue.popleft()
        # inference detection
        with stop_watch.timeit('Det'):
            mmdet_results = inference_detector(det_model, frame)

        t_info = stop_watch.report_strings()
        with det_result_queue_mutex:
            det_result_queue.append((ts_input, frame, t_info, mmdet_results))

        # limit the inference FPS
        _ts = time.time()
        if _ts_last is not None and _ts - _ts_last < min_interval:
            time.sleep(min_interval - _ts + _ts_last)
        _ts_last = time.time()


def inference_pose():
    print('Thread "pose" started')
    stop_watch = StopWatch(window=10)

    while True:
        while len(det_result_queue) < 1:
            time.sleep(0.001)
        with det_result_queue_mutex:
            ts_input, frame, t_info, mmdet_results = det_result_queue.popleft()

        pose_results_list = []
        for model_info, pose_history, smoother in zip(pose_model_list,
                                                      pose_history_list,
                                                      pose_smoother_list):
            model_name = model_info['name']
            pose_model = model_info['model']
            cat_ids = model_info['cat_ids']
            pose_results_last = pose_history['pose_results_last']
            next_id = pose_history['next_id']

            with stop_watch.timeit(model_name):
                # process mmdet results
                det_results = process_mmdet_results(
                    mmdet_results,
                    class_names=det_model.CLASSES,
                    cat_ids=cat_ids)

                # inference pose model
                dataset_name = pose_model.cfg.data['test']['type']
                pose_results, _ = inference_top_down_pose_model(
                    pose_model,
                    frame,
                    det_results,
                    bbox_thr=args.det_score_thr,
                    format='xyxy',
                    dataset=dataset_name)

                pose_results, next_id = get_track_id(
                    pose_results,
                    pose_results_last,
                    next_id,
                    use_oks=False,
                    tracking_thr=0.3)

                if smoother:
                    pose_results = smoother.smooth(pose_results)

                pose_results_list.append(pose_results)

                # update pose history
                pose_history['pose_results_last'] = pose_results
                pose_history['next_id'] = next_id

        t_info += stop_watch.report_strings()
        with pose_result_queue_mutex:
            pose_result_queue.append((ts_input, t_info, pose_results_list))

        event_inference_done.set()


def display():
    print('Thread "display" started')
    stop_watch = StopWatch(window=10)

    # initialize result status
    ts_inference = None  # timestamp of the latest inference result
    fps_inference = 0.  # infenrece FPS
    t_delay_inference = 0.  # inference result time delay
    pose_results_list = None  # latest inference result
    t_info = []  # upstream time information (list[str])

    # initialize visualization and output
    sunglasses_img = None  # resource image for sunglasses effect
    text_color = (228, 183, 61)  # text color to show time/system information
    vid_out = None  # video writer

    # show instructions
    print('Keyboard shortcuts: ')
    print('"v": Toggle the visualization of bounding boxes and poses.')
    print('"s": Toggle the sunglasses effect.')
    print('"b": Toggle the bug-eye effect.')
    print('"Q", "q" or Esc: Exit.')

    while True:
        with stop_watch.timeit('_FPS_'):
            # acquire a frame from buffer
            ts_input, frame = frame_buffer.get()
            # input ending signal
            if ts_input is None:
                break

            img = frame

            # get pose estimation results
            if len(pose_result_queue) > 0:
                with pose_result_queue_mutex:
                    _result = pose_result_queue.popleft()
                    _ts_input, t_info, pose_results_list = _result

                _ts = time.time()
                if ts_inference is not None:
                    fps_inference = 1.0 / (_ts - ts_inference)
                ts_inference = _ts
                t_delay_inference = (_ts - _ts_input) * 1000

            # visualize detection and pose results
            if pose_results_list is not None:
                for model_info, pose_results in zip(pose_model_list,
                                                    pose_results_list):
                    pose_model = model_info['model']
                    bbox_color = model_info['bbox_color']

                    dataset_name = pose_model.cfg.data['test']['type']

                    # show pose results
                    if args.vis_mode == 1:
                        img = vis_pose_result(
                            pose_model,
                            img,
                            pose_results,
                            radius=4,
                            thickness=2,
                            dataset=dataset_name,
                            kpt_score_thr=1e7,
                            bbox_color=bbox_color)
                    elif args.vis_mode == 2:
                        img = vis_pose_result(
                            pose_model,
                            img,
                            pose_results,
                            radius=4,
                            thickness=2,
                            dataset=dataset_name,
                            kpt_score_thr=args.kpt_thr,
                            bbox_color=bbox_color)

                    # sunglasses effect
                    if args.sunglasses:
                        if dataset_name in {
                                'TopDownCocoDataset',
                                'TopDownCocoWholeBodyDataset'
                        }:
                            left_eye_idx = 1
                            right_eye_idx = 2
                        elif dataset_name == 'AnimalPoseDataset':
                            left_eye_idx = 0
                            right_eye_idx = 1
                        else:
                            raise ValueError(
                                'Sunglasses effect does not support'
                                f'{dataset_name}')
                        if sunglasses_img is None:
                            # The image attributes to:
                            # https://www.vecteezy.com/free-vector/glass
                            # Glass Vectors by Vecteezy
                            sunglasses_img = cv2.imread(
                                'demo/resources/sunglasses.jpg')
                        img = apply_sunglasses_effect(img, pose_results,
                                                      sunglasses_img,
                                                      left_eye_idx,
                                                      right_eye_idx)
                    # bug-eye effect
                    if args.bugeye:
                        if dataset_name in {
                                'TopDownCocoDataset',
                                'TopDownCocoWholeBodyDataset'
                        }:
                            left_eye_idx = 1
                            right_eye_idx = 2
                        elif dataset_name == 'AnimalPoseDataset':
                            left_eye_idx = 0
                            right_eye_idx = 1
                        else:
                            raise ValueError('Bug-eye effect does not support'
                                             f'{dataset_name}')
                        img = apply_bugeye_effect(img, pose_results,
                                                  left_eye_idx, right_eye_idx)

            # delay control
            if args.display_delay > 0:
                t_sleep = args.display_delay * 0.001 - (time.time() - ts_input)
                if t_sleep > 0:
                    time.sleep(t_sleep)
            t_delay = (time.time() - ts_input) * 1000

            # show time information
            t_info_display = stop_watch.report_strings()  # display fps
            t_info_display.append(f'Inference FPS: {fps_inference:>5.1f}')
            t_info_display.append(f'Delay: {t_delay:>3.0f}')
            t_info_display.append(
                f'Inference Delay: {t_delay_inference:>3.0f}')
            t_info_str = ' | '.join(t_info_display + t_info)
            cv2.putText(img, t_info_str, (20, 20), cv2.FONT_HERSHEY_DUPLEX,
                        0.3, text_color, 1)
            # collect system information
            sys_info = [
                f'RES: {img.shape[1]}x{img.shape[0]}',
                f'Buffer: {frame_buffer.qsize()}/{frame_buffer.maxsize}'
            ]
            if psutil_proc is not None:
                sys_info += [
                    f'CPU: {psutil_proc.cpu_percent():.1f}%',
                    f'MEM: {psutil_proc.memory_percent():.1f}%'
                ]
            sys_info_str = ' | '.join(sys_info)
            cv2.putText(img, sys_info_str, (20, 40), cv2.FONT_HERSHEY_DUPLEX,
                        0.3, text_color, 1)

            # save the output video frame
            if args.out_video_file is not None:
                if vid_out is None:
                    fourcc = cv2.VideoWriter_fourcc(*'mp4v')
                    fps = args.out_video_fps
                    frame_size = (img.shape[1], img.shape[0])
                    vid_out = cv2.VideoWriter(args.out_video_file, fourcc, fps,
                                              frame_size)

                vid_out.write(img)

            # display
            cv2.imshow('mmpose webcam demo', img)
            keyboard_input = cv2.waitKey(1)
            if keyboard_input in (27, ord('q'), ord('Q')):
                break
            elif keyboard_input == ord('s'):
                args.sunglasses = not args.sunglasses
            elif keyboard_input == ord('b'):
                args.bugeye = not args.bugeye
            elif keyboard_input == ord('v'):
                args.vis_mode = (args.vis_mode + 1) % 3

    cv2.destroyAllWindows()
    if vid_out is not None:
        vid_out.release()
    event_exit.set()


def main():
    global args
    global frame_buffer
    global input_queue, input_queue_mutex
    global det_result_queue, det_result_queue_mutex
    global pose_result_queue, pose_result_queue_mutex
    global det_model, pose_model_list, pose_history_list
    global event_exit, event_inference_done
    global pose_smoother_list

    args = parse_args()

    assert has_mmdet, 'Please install mmdet to run the demo.'
    assert args.det_config is not None
    assert args.det_checkpoint is not None

    # build detection model
    det_model = init_detector(
        args.det_config, args.det_checkpoint, device=args.device.lower())

    # build pose models
    pose_model_list = []
    if args.enable_human_pose:
        pose_model = init_pose_model(
            args.human_pose_config,
            args.human_pose_checkpoint,
            device=args.device.lower())
        model_info = {
            'name': 'HumanPose',
            'model': pose_model,
            'cat_ids': args.human_det_ids,
            'bbox_color': (148, 139, 255),
        }
        pose_model_list.append(model_info)
    if args.enable_animal_pose:
        pose_model = init_pose_model(
            args.animal_pose_config,
            args.animal_pose_checkpoint,
            device=args.device.lower())
        model_info = {
            'name': 'AnimalPose',
            'model': pose_model,
            'cat_ids': args.animal_det_ids,
            'bbox_color': 'cyan',
        }
        pose_model_list.append(model_info)

    # store pose history for pose tracking
    pose_history_list = []
    for _ in range(len(pose_model_list)):
        pose_history_list.append({'pose_results_last': [], 'next_id': 0})

    # build pose smoother for temporal refinement
    pose_smoother_list = []
    for _ in range(len(pose_model_list)):
        if args.smooth:
            smoother = Smoother(
                filter_cfg=args.smooth_filter_cfg, keypoint_dim=2)
        else:
            smoother = None
        pose_smoother_list.append(smoother)

    # frame buffer
    if args.buffer_size > 0:
        buffer_size = args.buffer_size
    else:
        # infer buffer size from the display delay time
        # assume that the maximum video fps is 30
        buffer_size = round(30 * (1 + max(args.display_delay, 0) / 1000.))
    frame_buffer = Queue(maxsize=buffer_size)

    # queue of input frames
    # element: (timestamp, frame)
    input_queue = deque(maxlen=1)
    input_queue_mutex = Lock()

    # queue of detection results
    # element: tuple(timestamp, frame, time_info, det_results)
    det_result_queue = deque(maxlen=1)
    det_result_queue_mutex = Lock()

    # queue of detection/pose results
    # element: (timestamp, time_info, pose_results_list)
    pose_result_queue = deque(maxlen=1)
    pose_result_queue_mutex = Lock()

    try:
        event_exit = Event()
        event_inference_done = Event()
        t_input = Thread(target=read_camera, args=())
        t_det = Thread(target=inference_detection, args=(), daemon=True)
        t_pose = Thread(target=inference_pose, args=(), daemon=True)

        t_input.start()
        t_det.start()
        t_pose.start()

        # run display in the main thread
        display()
        # join the input thread (non-daemon)
        t_input.join()

    except KeyboardInterrupt:
        pass


if __name__ == '__main__':
    main()
